# Copyright 2016-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.tofu import scenario, expected, Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import XPath
from reahl.component.exceptions import IsInstance

from reahl.web.fw import Bookmark
from reahl.web.bootstrap.ui import A, Div, P, TextNode, Span
from reahl.web.bootstrap.forms import Form
from reahl.web.bootstrap.navbar import Navbar, NavbarLayout, ResponsiveLayout, CollapseToggle
from reahl.web.bootstrap.navs import Nav

from reahl.component.exceptions import ProgrammerError

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class NavbarFixture(Fixture):

    def new_navbar(self):
        return Navbar(self.web_fixture.view)

    def new_navbar_with_layout(self):
        return self.navbar.use_layout(NavbarLayout())

    def new_bookmarks(self):
        return [Bookmark('', '/one', 'One')]

    def new_nav(self):
        return Nav(self.web_fixture.view).with_bookmarks(self.bookmarks)

    def new_form(self):
        return Form(self.web_fixture.view, 'myform')

    def new_textnode(self):
        return TextNode(self.web_fixture.view, 'mytext')


@with_fixtures(WebFixture, NavbarFixture)
def test_navbar_basics(web_fixture, navbar_fixture):
    """A typical Navbar is created by using its layout to add some brand text, a nav and form in it."""

    fixture = navbar_fixture

    navbar = Navbar(web_fixture.view).use_layout(NavbarLayout())

    navbar.layout.set_brand_text('Brandy')
    navbar.layout.add(fixture.nav)
    navbar.layout.add(fixture.form)

    [brand, nav, form] = navbar.children[0].children
    [ul] = nav.children

    # The Navbar itself
    assert navbar.children[0].tag_name == 'nav'
    assert 'navbar' in navbar.children[0].get_attribute('class').split(' ')

    # The added contents
    assert isinstance(brand, A)
    assert brand.href.path == '/'
    assert 'navbar-brand' in brand.get_attribute('class')

    assert 'navbar-nav' in ul.get_attribute('class')

    assert isinstance(form, Form)
    assert 'form-inline' in form.get_attribute('class')


@uses(web_fixture=WebFixture)
class LayoutScenarios(Fixture):

    @scenario
    def fixed_top(self):
        self.layout = NavbarLayout(fixed_to='fixed-top')
        self.expected_css_class = 'fixed-top'

    @scenario
    def fixed_bottom(self):
        self.layout = NavbarLayout(fixed_to='fixed-bottom')
        self.expected_css_class = 'fixed-bottom'

    @scenario
    def sticky_top(self):
        self.layout = NavbarLayout(fixed_to='sticky-top')
        self.expected_css_class = 'sticky-top'

    @scenario
    def default(self):
        self.layout = NavbarLayout()
        self.expected_css_class = None


@with_fixtures(WebFixture, LayoutScenarios)
def test_navbar_can_have_layout(web_fixture, layout_scenarios):
    """NavbarLayout is used to define the placement of a Navbar."""

    fixture = layout_scenarios

    widget = Navbar(web_fixture.view).use_layout(fixture.layout)

    [navbar] = widget.children
    all_classes = ['fixed-bottom', 'fixed-top', 'sticky-top']
    if fixture.expected_css_class:
        assert fixture.expected_css_class in navbar.get_attribute('class').split(' ')

    for not_expected_class in [i for i in all_classes if i != fixture.expected_css_class]:
        assert not_expected_class not in navbar.get_attribute('class').split(' ')


@with_fixtures(WebFixture)
def test_customised_colour_scheme(web_fixture):
    """A ColourScheme is used to determine link colours and/or optionally a standard bootstrap background color."""

    layout = NavbarLayout(colour_theme='light', bg_scheme='dark')
    widget = Navbar(web_fixture.view).use_layout(layout)

    [navbar] = widget.children

    assert 'navbar-light' in navbar.get_attribute('class')
    assert 'bg-dark' in navbar.get_attribute('class')


@with_fixtures(WebFixture, NavbarFixture)
def test_adding_brand_widget(web_fixture, navbar_fixture):
    """Brand content can also be added as a Widget, instead of only text."""

    navbar_widget = navbar_fixture.navbar.use_layout(NavbarLayout())
    custom_brand = Div(web_fixture.view)
    navbar_widget.layout.set_brand(custom_brand)

    [navbar] = navbar_widget.children
    [actual_brand_widget] = navbar.children

    assert actual_brand_widget is custom_brand
    assert 'navbar-brand' in actual_brand_widget.get_attribute('class')


@with_fixtures(WebFixture, NavbarFixture)
def test_adding_other_than_form_or_nav_is_not_allowed(web_fixture, navbar_fixture):
    """Only Navs, Forms and Text may be added to a NavbarLayout."""

    navbar = navbar_fixture.navbar.use_layout(NavbarLayout())
    not_a_form_or_nav = Div(web_fixture.view)

    with expected(IsInstance):
        navbar.layout.add(not_a_form_or_nav)

    # Case: Form
    navbar = navbar_fixture.new_navbar().use_layout(NavbarLayout())
    navbar.layout.add(navbar_fixture.form)
    [added_widget] = navbar.children[0].children
    assert added_widget is navbar_fixture.form
    assert 'form-inline' in navbar_fixture.form.get_attribute('class')

    # Case: Nav
    navbar = navbar_fixture.new_navbar().use_layout(NavbarLayout())
    assert 'navbar-nav' not in navbar_fixture.nav.html_representation.get_attribute('class').split(' ')
    navbar.layout.add(navbar_fixture.nav)
    [added_widget] = navbar.children[0].children
    assert added_widget is navbar_fixture.nav
    assert 'navbar-nav' in navbar_fixture.nav.html_representation.get_attribute('class').split(' ')

    # Case: Text
    navbar = navbar_fixture.new_navbar().use_layout(NavbarLayout())
    navbar.layout.add(navbar_fixture.textnode)
    [added_widget] = navbar.children[0].children
    [textnode] = added_widget.children
    assert isinstance(added_widget, Span)
    assert textnode is navbar_fixture.textnode
    assert 'navbar-text' in added_widget.get_attribute('class').split(' ')



@with_fixtures(WebFixture, NavbarFixture)
def test_navbar_with_centered_contents(web_fixture, navbar_fixture):
    """Contents of a Navbar appears centered when center_contents is set to True"""

    navbar_widget = navbar_fixture.navbar
    navbar_widget.use_layout(NavbarLayout(center_contents=True))
    navbar_widget.layout.set_brand_text('Brandy') #adding something to illustrate the structure change

    [navbar] = navbar_widget.children
    [centering_div] = navbar.children
    [brand_widget] = centering_div.children

    assert 'container' in centering_div.get_attribute('class')
    assert 'navbar-brand' in brand_widget.get_attribute('class')


@uses(web_fixture=WebFixture)
class NavbarToggleFixture(Fixture):

    def is_expanded(self, locator):
        return self.web_fixture.driver_browser.is_visible(locator) and \
               self.web_fixture.driver_browser.does_element_have_attribute(locator, 'class', value='collapse show')

    def panel_is_visible(self):
        return self.web_fixture.driver_browser.is_visible(XPath.paragraph().including_text('Peek-A-Boo'))

    def panel_is_expanded(self):
        return self.is_expanded(XPath.paragraph().including_text('Peek-A-Boo'))

    def xpath_to_locate_toggle(self):
        return XPath('//span[contains(@class,"navbar-toggler-icon")]')

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Div):
            def __init__(self, view):
                super().__init__(view)

                navbar = Navbar(view)
                navbar.use_layout(NavbarLayout(colour_theme='dark', bg_scheme='dark'))
                fixture.element_to_collapse = P(view, text='Peek-A-Boo', css_id='my_id')
                navbar.layout.add_toggle(fixture.element_to_collapse)

                self.add_child(fixture.element_to_collapse)
                self.add_child(navbar)
        return MainWidget


@with_fixtures(WebFixture, NavbarFixture)
def test_navbar_toggle_basics(web_fixture, navbar_fixture):
    """You can add a toggle to the navbar that hides another element on the page."""

    element_to_collapse = Div(web_fixture.view, css_id='my_id')
    toggle = navbar_fixture.navbar_with_layout.layout.add_toggle(element_to_collapse)

    [toggle_span] = toggle.children

    assert isinstance(toggle_span, Span)
    assert 'navbar-toggler-icon' in toggle_span.get_attribute('class')
    assert 'navbar-toggler' in toggle.get_attribute('class').split()


@with_fixtures(WebFixture, NavbarFixture)
def test_navbar_toggle_customised(web_fixture, navbar_fixture):
    """The text on a toggle that hides an element is customisable"""

    element_to_collapse = Div(web_fixture.view, css_id='my_id')
    toggle = navbar_fixture.navbar_with_layout.layout.add_toggle(element_to_collapse, text='≎')

    [toggle_text_node] = toggle.children

    assert('≎' == toggle_text_node.value)


@with_fixtures(WebFixture, NavbarToggleFixture)
def test_navbar_toggle_collapses_html_element(web_fixture, navbar_toggle_fixture):
    """You can add a toggle to the navbar that hides another element on the page."""

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=navbar_toggle_fixture.MainWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')


    #case: by default, the element to hide is not visible
    assert browser.wait_for_not(navbar_toggle_fixture.panel_is_visible)

    #case: clicking on the toggle, causes the panel to appear
    browser.click(navbar_toggle_fixture.xpath_to_locate_toggle())
    browser.wait_for(navbar_toggle_fixture.panel_is_expanded)

    # case: clicking on the toggle again, causes the panel to disappear
    browser.click(navbar_toggle_fixture.xpath_to_locate_toggle())
    browser.wait_for_not(navbar_toggle_fixture.panel_is_visible)



@with_fixtures(WebFixture, NavbarFixture)
def test_navbar_toggle_requires_target_id(web_fixture, navbar_fixture):
    """To be able to hide an element, it is required to have an id"""

    navbar = navbar_fixture.navbar
    navbar.use_layout(NavbarLayout())
    element_without_id = P(web_fixture.view, text='Peek-A-Boo')

    with expected(ProgrammerError, test='.*has no css_id set.*'):
        navbar.layout.add_toggle(element_without_id)


@with_fixtures(WebFixture, NavbarFixture)
def test_responsive_navbar_basics(web_fixture, navbar_fixture):
    """A ResponsiveLayout hides its Navbar when the viewport becomes smaller than a given device size"""

    navbar_widget = navbar_fixture.navbar
    navbar_widget.set_id('my_navbar_id')
    navbar_widget.use_layout(ResponsiveLayout('sm'))

    [navbar] = navbar_widget.children
    [toggle, collapse_div] = navbar.children

    assert 'navbar-toggler' in toggle.get_attribute('class')
    assert 'button' in toggle.get_attribute('type')
    assert 'collapse' in toggle.get_attribute('data-toggle')
    assert 'my_navbar_id' in toggle.get_attribute('data-target')

    assert 'my_navbar_id' in collapse_div.get_attribute('id')
    assert 'collapse' in collapse_div.get_attribute('class').split()
    assert 'navbar-collapse' in collapse_div.get_attribute('class').split()

    assert 'navbar-expand-xs' in navbar.get_attribute('class')


class ToggleAlignmentScenarios(Fixture):
    @scenario
    def aligned_right(self):
        self.layout = ResponsiveLayout('sm', align_toggle_left=False)
        self.expected_order = [A, CollapseToggle, Div]

    @scenario
    def aligned_left(self):
        self.layout = ResponsiveLayout('sm', align_toggle_left=True)
        self.expected_order = [CollapseToggle, A, Div]


@with_fixtures(NavbarFixture, ToggleAlignmentScenarios)
def test_responsive_navbar_toggle_alignment(navbar_fixture, toggle_alignment_fixture):
    """A ResponsiveLayout hides its Navbar when the viewport becomes smaller than a given device size"""

    navbar_widget = navbar_fixture.navbar
    navbar_widget.set_id('my_navbar_id')
    navbar_widget.use_layout(toggle_alignment_fixture.layout)
    navbar_widget.layout.set_brand_text('brand')

    [navbar] = navbar_widget.children

    assert [c.__class__ for c in navbar.children] == toggle_alignment_fixture.expected_order


class BrandCollapseScenarios(Fixture):

    @scenario
    def brand_does_not_collapse_with_content(self):
        self.brand_collapse = False

    @scenario
    def brand_does_collapse_with_content(self):
        self.brand_collapse = True


@with_fixtures(WebFixture, NavbarFixture, BrandCollapseScenarios)
def test_brand_may_be_collapsed_with_expandable_content(web_fixture, navbar_fixture, brand_collapse_fixture):
    """Brands may be collapse along with the other content."""

    responsive_navbar = navbar_fixture.navbar
    responsive_navbar.set_id('my_navbar_id')
    responsive_navbar.use_layout(ResponsiveLayout('md', collapse_brand_with_content=brand_collapse_fixture.brand_collapse))
    responsive_navbar.layout.add(navbar_fixture.nav)

    brand_widget = Div(web_fixture.view)
    responsive_navbar.layout.set_brand(brand_widget)

    if brand_collapse_fixture.brand_collapse:
        [toggle, collapse_div] = responsive_navbar.children[0].children
        [brand, navbar_nav] = collapse_div.children
    else:
        [brand, toggle, collapse_div] = responsive_navbar.children[0].children
        [navbar_nav] = collapse_div.children

    assert brand is brand_widget
    assert 'navbar-collapse' in collapse_div.get_attribute('class')
    assert 'navbar-nav' in navbar_nav.html_representation.get_attribute('class')
    assert 'navbar-toggler' in toggle.get_attribute('class')



# Copyright 2021, 2022, 2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
import datetime
import base64

from selenium.common.exceptions import TimeoutException

from reahl.tofu import Fixture, scenario, expected, uses, NoException
from reahl.tofu.pytestsupport import with_fixtures

from reahl.component.context import  ExecutionContext
from reahl.component.modelinterface import Action, Event, ExposedNames, ChoiceField, Choice, IntegerField
from reahl.web.ui import Form, ButtonInput, FormLayout, SelectInput, P
from reahl.web.csrf import ExpiredCSRFToken, InvalidCSRFToken, CSRFToken
from reahl.browsertools.browsertools import XPath

from reahl.web_dev.fixtures import WebFixture



class CSRFFixture(Fixture):
    def new_MyForm(self):
        class MyForm(Form):
            def __init__(self, view):
                super().__init__(view, 'myform')
                self.enable_refresh()
                self.choice = 1
                self.use_layout(FormLayout())
                if self.exception:
                    self.layout.add_alert_for_domain_exception(self.exception)
                self.layout.add_input(SelectInput(self, self.fields.choice, refresh_widget=self))
                self.define_event_handler(self.events.submit_break)
                self.add_child(ButtonInput(self, self.events.submit_break))
                self.add_child(P(view, 'Chosen: %s' % self.choice))

            events = ExposedNames()
            events.submit_break = lambda i: Event(label='Submit')

            fields = ExposedNames()
            fields.choice = lambda i: ChoiceField([Choice(1, IntegerField(label='1')),
                                                   Choice(2, IntegerField(label='2'))])


        return MyForm

    def set_csrf_token_in_rendered_form(self, browser, value):
        browser.execute_script('$("#id-myform-_reahl_csrf_token").attr("value", "%s")' % value)

    def get_csrf_token_in_rendered_form(self, browser):
        return browser.execute_script('return $("#id-myform-_reahl_csrf_token").attr("value")')

    def set_csrf_token_in_rendered_form_to_expired(self, browser):
        valid_token = self.get_csrf_token_in_rendered_form(browser)
        reconstructed_token = CSRFToken.from_coded_string(valid_token)
        allowed_timeout = ExecutionContext.get_context().config.web.csrf_timeout_seconds
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        stale_time = now - datetime.timedelta(seconds=allowed_timeout + 1)
        reconstructed_token.timestamp = stale_time.timestamp()
        stale_token_string = reconstructed_token.as_signed_string()
        self.set_csrf_token_in_rendered_form(browser, stale_token_string)


@with_fixtures(WebFixture, CSRFFixture)
def test_submit_form_with_invalid_csrf_token(web_fixture, csrf_fixture):
    """A Form cannot be submitted without the original CSRF token proving that it was rendered by us originally."""
    fixture = csrf_fixture

    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.MyForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    browser.open('/')
    fixture.set_csrf_token_in_rendered_form(browser, 'invalid csrf token value')
    assert not browser.is_element_present(XPath.heading(1).with_text("403 Forbidden"))
    browser.click(XPath.button_labelled('Submit'))
    assert browser.is_element_present(XPath.heading(1).with_text("403 Forbidden"))


@with_fixtures(WebFixture, CSRFFixture)
def test_submit_form_with_expired_csrf_token(web_fixture, csrf_fixture):
    """A form submitted with a valid expired token, shows a validation exception. After refresh, a new token is received and submit works."""
    fixture = csrf_fixture

    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.MyForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    browser.open('/')
    select_widget_path = XPath.select_labelled('choice')
    assert browser.get_value(select_widget_path) == '1'
    browser.select(select_widget_path, '2')

    wsgi_app.config.web.csrf_timeout_seconds = 0.5
    time.sleep(wsgi_app.config.web.csrf_timeout_seconds+0.1)
    browser.click(XPath.button_labelled('Submit'))
    wsgi_app.config.web.csrf_timeout_seconds = 300
    error_message = XPath.paragraph().including_text('This page has expired. For security reasons, please review your input and retry.')
    assert browser.is_element_present(error_message)

    #case: submit again should work now - new csrftoken received in GET
    browser.click(XPath.button_labelled('Submit'))
    assert not browser.is_element_present(error_message)


@with_fixtures(WebFixture, CSRFFixture)
def test_refresh_widget_with_expired_csrf_token(web_fixture, csrf_fixture):
    """An Ajax request done with a valid expired token, shows a validation exception. After refresh, a new token is received and submit works."""
    fixture = csrf_fixture

    wsgi_app = web_fixture.new_wsgi_app(child_factory=fixture.MyForm.factory(), enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    wsgi_app.config.reahlsystem.debug = False #causes redirect to the error page and not stacktrace

    browser.open('/')

    select_widget_path = XPath.select_labelled('choice')
    assert browser.get_value(select_widget_path) == '1'
    wsgi_app.config.web.csrf_timeout_seconds = 0.5
    time.sleep(wsgi_app.config.web.csrf_timeout_seconds+0.1)
    browser.select(select_widget_path, '2', wait_for_ajax=False)
    wsgi_app.config.web.csrf_timeout_seconds = 300

    with browser.alert_expected() as js_alert:
        assert js_alert.text == 'This page has expired. For security reasons, please review your input and retry.'
        js_alert.accept()

    #case: new csrftoken received in GET
    assert browser.get_value(select_widget_path) == '1'
    assert browser.wait_for_element_not_present(XPath.paragraph().with_text('Chosen: 2'))
    browser.select(select_widget_path, '2')

    assert browser.wait_for_element_present(XPath.paragraph().with_text('Chosen: 2'))
    assert browser.wait_for_not(browser.is_alert_present)


@with_fixtures(WebFixture)
def test_check_csrf_token_match(web_fixture):
    """A CSRFToken is transformed into a signed string, with timestamp, and can be reconstructed and matched with a reconstructed token."""
    token = CSRFToken(value='hello world')
    csrf_token = token.as_signed_string()

    assert len(csrf_token.split(':')) == 3

    reconstructed_token = CSRFToken.from_coded_string(csrf_token)
    assert reconstructed_token.value == 'hello world'
    assert reconstructed_token.matches(token)


@with_fixtures(WebFixture)
def test_csrf_fiddled_value(web_fixture):
    """A token with signature that does not match its contents is invalid."""
    token = CSRFToken(value='hello world')
    csrf_token = token.as_signed_string()
    value, timestamp, signature = csrf_token.split(":")

    crafted_value = base64.urlsafe_b64encode('new world'.encode()).decode('utf-8')
    crafted_token_string = ':'.join([crafted_value, timestamp, signature])
    with expected(InvalidCSRFToken):
        CSRFToken.from_coded_string(crafted_token_string)


@with_fixtures(WebFixture)
def test_csrf_malformed_token(web_fixture):
    """A malformed token is invalid."""
    with expected(InvalidCSRFToken):
        CSRFToken.from_coded_string("someting without delimeters")


@with_fixtures(WebFixture)
def test_csrf_with_invalid_timestamp(web_fixture):
    """A Token with a timestamp in the future is invalid."""
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    seconds_into_future = 300
    future_time = now + datetime.timedelta(seconds=seconds_into_future)
    future_token = CSRFToken(timestamp=future_time)
    with expected(InvalidCSRFToken):
        CSRFToken.from_coded_string(future_token.as_signed_string())



# Copyright 2013-2024 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import json
from webob import Response

from sqlalchemy import Column, UnicodeText, Integer

from reahl.stubble import stubclass, CallMonitor, replaced
from reahl.tofu import scenario, expected, Fixture, uses, NoException
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser, XPath
from reahl.component.exceptions import DomainException, ProgrammerError
from reahl.web.fw import CheckedRemoteMethod, JsonResult, MethodResult, RemoteMethod, Widget, WidgetResult
from reahl.web.ui import Div
from reahl.component.modelinterface import Field, IntegerField

from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture
from reahl.sqlalchemysupport import Base, Session
from reahl.dev.fixtures import ReahlSystemFixture
from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class RemoteMethodFixture(Fixture):

    def new_widget_factory(self, remote_method=None):
        remote_method = remote_method or self.remote_method

        @stubclass(Widget)
        class WidgetWithRemoteMethod(Widget):
            def __init__(self, view):
                super().__init__(view)
                view.add_resource(remote_method)

        return WidgetWithRemoteMethod.factory()

    def new_wsgi_app(self, remote_method=None, enable_js=False):
        remote_method = remote_method or self.remote_method
        return self.web_fixture.new_wsgi_app(child_factory=self.new_widget_factory(remote_method=remote_method), enable_js=enable_js)


@with_fixtures(WebFixture, RemoteMethodFixture)
def test_remote_methods(web_fixture, remote_method_fixture):
    """A RemoteMethod is a SubResource representing a method on the server side which can be invoked via POSTing to an URL."""

    fixture = remote_method_fixture

    def callable_object():
        return 'value returned from method'

    encoding = 'koi8_r'  # Deliberate
    remote_method = RemoteMethod(web_fixture.view, 'amethod', callable_object, MethodResult(mime_type='ttext/hhtml', encoding=encoding), disable_csrf_check=True)

    wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    # By default you cannot GET, since the method is not immutable
    browser.open('/_amethod_method', status=405)

    # POSTing to the URL, returns the result of the method
    browser.post('/_amethod_method', {})
    assert browser.raw_html == 'value returned from method'
    assert browser.last_response.charset == encoding
    assert browser.last_response.content_type == 'ttext/hhtml'


@with_fixtures(WebFixture, RemoteMethodFixture)
def test_remote_methods_via_ajax(web_fixture, remote_method_fixture):
    """A RemoteMethod can be called via AJAX with CSRF protection built-in."""

    fixture = remote_method_fixture

    def callable_object():
        return 'value returned from method'

    remote_method = RemoteMethod(web_fixture.view, 'amethod', callable_object, MethodResult(), disable_csrf_check=False)

    wsgi_app = fixture.new_wsgi_app(remote_method=remote_method, enable_js=True)
    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser

    # Case: using jquery to POST to the method includes the necessary xsrf info automatically
    browser.open('/')
    browser.execute_script('$.post("/_amethod_method", success=function(data){ $("body").attr("data-result", data) })')
    browser.wait_for_element_present(XPath('body/@data-result'))
    results = browser.execute_script('return $("body").attr("data-result")')
    assert results == 'value returned from method'

    # Case: POSTing without a csrf token breaks
    browser = Browser(wsgi_app)
    browser.post('/_amethod_method', {}, status=403)


@with_fixtures(WebFixture, RemoteMethodFixture)
def test_exception_handling(web_fixture, remote_method_fixture):
    """The RemoteMethod sends back the str() of an exception raised for the specified exception class."""


    def fail():
        raise Exception('I failed')
    remote_method = RemoteMethod(web_fixture.view, 'amethod', fail, MethodResult(catch_exception=Exception), disable_csrf_check=True)

    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    browser.post('/_amethod_method', {})
    assert browser.raw_html == 'I failed'


@with_fixtures(WebFixture, RemoteMethodFixture)
def test_idempotent_remote_methods(web_fixture, remote_method_fixture):
    """A RemoteMethod that is idempotent is accessible via GET (instead of POST)."""

    def callable_object():
        return 'value returned from method'
    remote_method = RemoteMethod(web_fixture.view, 'amethod', callable_object, MethodResult(), idempotent=True, disable_csrf_check=True)

    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    # GET, since the method is idempotent
    browser.open('/_amethod_method')
    assert browser.raw_html == 'value returned from method'

    # POSTing to the URL, is not supported
    browser.post('/_amethod_method', {}, status=405)


@with_fixtures(WebFixture, RemoteMethodFixture, SqlAlchemyFixture)
def test_immutable_remote_methods(web_fixture, remote_method_fixture, sql_alchemy_fixture):
    """The database is always rolled back at the end of an immutable RemoteMethod."""

    class TestObject(Base):
        __tablename__ = 'test_remotemethods_test_object'
        id = Column(Integer, primary_key=True)
        name = Column(UnicodeText)

    with sql_alchemy_fixture.persistent_test_classes(TestObject):
        def callable_object():
            Session.add(TestObject(name='new object'))
            assert Session.query(TestObject).count() == 1
            return 'value returned from method'
        
        remote_method = RemoteMethod(web_fixture.view, 'amethod', callable_object, MethodResult(), immutable=True, disable_csrf_check=True)

        assert remote_method.idempotent  # Immutable methods are idempotent

        wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)

        browser.open('/_amethod_method')
        assert browser.raw_html == 'value returned from method'

        # The database is rolled back to ensure immutability
        assert Session.query(TestObject).count() == 0


class ArgumentScenarios(Fixture):
    @scenario
    def get(self):
        self.idempotent = True
    @scenario
    def post(self):
        self.idempotent = False


@with_fixtures(WebFixture, RemoteMethodFixture, ArgumentScenarios)
def test_arguments_to_remote_methods(web_fixture, remote_method_fixture, argument_scenarios):
    """A RemoteMethod can get arguments from a query string or submitted form values, depending on the scenario."""

    fixture = argument_scenarios

    def callable_object(**kwargs):
        fixture.method_kwargs = kwargs
        return ''

    remote_method = RemoteMethod(web_fixture.view, 'amethod', callable_object, MethodResult(), idempotent=fixture.idempotent, disable_csrf_check=True)

    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    kwargs_sent = {'a':'AAA', 'b':'BBB'}
    if fixture.idempotent:
        browser.open('/_amethod_method?a=AAA&b=BBB')
    else:
        browser.post('/_amethod_method', kwargs_sent)
    assert fixture.method_kwargs == kwargs_sent


@with_fixtures(WebFixture, RemoteMethodFixture, ArgumentScenarios)
def test_checked_arguments(web_fixture, remote_method_fixture, argument_scenarios):
    """A CheckedRemoteMethod checks and marshalls its parameters using Fields."""

    fixture = argument_scenarios

    def callable_object(anint=None, astring=None):
        fixture.method_kwargs = {'anint': anint, 'astring': astring}
        return ''

    remote_method = CheckedRemoteMethod(web_fixture.view, 'amethod', callable_object, MethodResult(),
                                        idempotent=fixture.idempotent,
                                        anint=IntegerField(),
                                        astring=Field(),
                                        disable_csrf_check=True)

    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    if fixture.idempotent:
        browser.open('/_amethod_method?anint=5&astring=SupercalifraGilisticexpialidocious')
    else:
        browser.post('/_amethod_method', {'anint':'5', 'astring':'SupercalifraGilisticexpialidocious'})
    assert fixture.method_kwargs == {'anint':5, 'astring':'SupercalifraGilisticexpialidocious'}


@uses(web_fixture=WebFixture)
class ResultScenarios(Fixture):
    @scenario
    def json(self):
        self.method_result = JsonResult(IntegerField(), catch_exception=Exception)
        self.value_to_return = 1
        self.expected_response = '1'
        self.exception_response = '"exception text"'
        self.expected_charset = self.method_result.encoding
        self.expected_content_type = 'application/json'
        self.results_match = lambda x, y: x == y

    @stubclass(Widget)
    class WidgetStub(Widget):
        css_id = 'someid'
        def render_contents(self): return '<the widget contents>'
        def get_contents_js(self, context=None): return ['some', 'some', 'javascript']

    @scenario
    def widget_as_json(self):
        self.method_result = WidgetResult([self.WidgetStub(self.web_fixture.view)])
        [self.web_fixture.view.page] = self.method_result.result_widgets

        self.value_to_return = 'ignored in this case'
        self.expected_response = {'result': {"someid": '<the widget contents><script type="text/javascript">javascriptsome</script>'},
                                  'success': True, 'exception': ''}
        self.expected_charset = self.method_result.encoding
        self.expected_content_type = 'application/json'
        def results_match(expected, actual):
            return json.loads(actual) == expected
        self.results_match = results_match


@with_fixtures(WebFixture, RemoteMethodFixture, ResultScenarios)
def test_different_kinds_of_result(web_fixture, remote_method_fixture, result_scenarios):
    """Different kinds of MethodResult can be specified for a method."""

    fixture = result_scenarios

    def callable_object():
        return fixture.value_to_return
    remote_method = RemoteMethod(web_fixture.view, 'amethod', callable_object, default_result=fixture.method_result, disable_csrf_check=True)

    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    browser.post('/_amethod_method', {})
    assert fixture.results_match(fixture.expected_response, browser.raw_html)
    assert browser.last_response.charset == fixture.expected_charset
    assert browser.last_response.content_type == fixture.expected_content_type


@with_fixtures(WebFixture, RemoteMethodFixture, ResultScenarios.json)
def test_exception_handling_for_json(web_fixture, remote_method_fixture, json_result_scenario):
    """How exceptions are handled with JsonResult."""

    fixture = json_result_scenario

    def fail():
        raise Exception('exception text')
    remote_method = RemoteMethod(web_fixture.view, 'amethod', fail, default_result=fixture.method_result, disable_csrf_check=True)

    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=remote_method)
    browser = Browser(wsgi_app)

    browser.post('/_amethod_method', {})
    assert browser.raw_html == fixture.exception_response
    assert browser.last_response.charset == fixture.expected_charset
    assert browser.last_response.content_type == fixture.expected_content_type



@uses(web_fixture=WebFixture)
class RegenerateMethodResultScenarios(Fixture):
    method_called = 0
    def new_method_result(self):
        fixture = self
        @stubclass(MethodResult)
        class ResultThatDependsOnExecutionOfMethod(MethodResult):
            def create_response(self, return_value):
                return Response(body='success: method was called %s times' % fixture.method_called)

            def create_exception_response(self, exception):
                return Response(body='exception: method was called %s times' % fixture.method_called)
        return ResultThatDependsOnExecutionOfMethod(catch_exception=DomainException, replay_request=True)

    def new_remote_method(self):
        fixture = self
        def callable_to_call():
            fixture.method_called += 1
            if fixture.exception:
                raise DomainException('ex')
        return RemoteMethod(self.web_fixture.view, 'amethod', callable_to_call, self.method_result, immutable=False, disable_csrf_check=True)

    @scenario
    def success(self):
        self.exception = False
        self.expected_response = 'success: method was called 1 times'

    @scenario
    def exception(self):
        self.exception = True
        self.expected_response = 'exception: method was called 1 times'


@with_fixtures(ReahlSystemFixture, WebFixture, RemoteMethodFixture, RegenerateMethodResultScenarios)
def test_regenerating_method_results(reahl_system_fixture, web_fixture,
                                     remote_method_fixture, regenerate_method_result_scenarios):
    """If a MethodResult is set up to replay_request=True, the view it is part of (and thus itself) is recreated
       before the (new incarnation of the) MethodResult generates its actual response. Replaying the request means recreating
       all Widgets on the current View as well as the MethodResult itself. The construction of any of these
       objects may happen differently because of the changes made during the RemoteMethod's execution. Replaying
       the request ensures that the MethodResult reflects such changes, yet ensures that the RemoteMethod
       is not executed twice.
    """


    wsgi_app = remote_method_fixture.new_wsgi_app(remote_method=regenerate_method_result_scenarios.remote_method)
    browser = Browser(wsgi_app)

    import sqlalchemy.orm 
    @stubclass(sqlalchemy.orm.Session)
    class TransactionStub:
        is_active=True
        def commit(self):pass
        def rollback(self):pass
    
    def wrapped_nested_transaction():
        return web_fixture.nested_transaction
    
    web_fixture.nested_transaction =  TransactionStub()
    with replaced(Session().begin_nested, wrapped_nested_transaction):
        with CallMonitor(web_fixture.nested_transaction.commit) as monitor:
            browser.post('/_amethod_method', {})
        assert browser.raw_html == regenerate_method_result_scenarios.expected_response
        assert monitor.times_called == 2


class WidgetResultScenarios(Fixture):
    changes_made = False
    def new_WidgetChildStub(self):
        @stubclass(Widget)
        class WidgetChildStub(Widget):
            def __init__(self, view, text):
                super().__init__(view)
                self.text = text
            def render(self): return '<%s>' % self.text
            def get_js(self, context=None): return ['js(%s)' % self.text]
        return WidgetChildStub

    def new_WidgetStub(self):
        fixture = self
        @stubclass(Widget)
        class WidgetStub(Widget):
            css_id = 'an_id'
            def __init__(self, view):
                super().__init__(view)
                message = 'changed contents' if fixture.changes_made else 'initial contents'
                self.add_child(fixture.WidgetChildStub(view, message))
        return WidgetStub

    def new_WidgetWithRemoteMethod(self):
        fixture = self
        @stubclass(Widget)
        class WidgetWithRemoteMethod(Widget):
            def __init__(self, view):
                super().__init__(view)
                method_result = WidgetResult([fixture.WidgetStub(view)])
                def change_something():
                    fixture.changes_made = True
                    if fixture.exception:
                        raise DomainException(message='breaking intentionally', handled_inline=fixture.handle_inline)
                remote_method = RemoteMethod(view, 'amethod', change_something, default_result=method_result,
                                             immutable=False, disable_csrf_check=True)
                view.add_resource(remote_method)
        return WidgetWithRemoteMethod

    @scenario
    def success(self):
        self.exception = False
        self.handle_inline = True
        self.expected_response = {'success': True,
                                  'exception': '',
                                  'result': {'an_id': '<changed contents><script type="text/javascript">js(changed contents)</script>'}}

    @scenario
    def exception(self):
        self.exception = True
        self.handle_inline = False
        self.expected_response = {'success': False,
                                  'exception': 'breaking intentionally',
                                  'result': {'an_id': '<changed contents><script type="text/javascript">js(changed contents)</script>'}}

    @scenario
    def exception_inline(self):
        self.exception = True
        self.handle_inline = True
        self.expected_response = {'success': False,
                                  'exception': '',
                                  'result': {'an_id': '<changed contents><script type="text/javascript">js(changed contents)</script>'}}

@with_fixtures(WebFixture, WidgetResultScenarios)
def test_widgets_that_change_during_method_processing(web_fixture, widget_result_scenarios):
    """The Widget rendered by WidgetResult reflects its Widget as it would have
       looked if it were constructed AFTER the changes effected by executing
       its RemoteMethod have been committed.
    """

    wsgi_app = web_fixture.new_wsgi_app(child_factory=widget_result_scenarios.WidgetWithRemoteMethod.factory())
    browser = Browser(wsgi_app)

    browser.post('/_amethod_method', {})
    json_response = json.loads(browser.raw_html)
    assert json_response == widget_result_scenarios.expected_response


@stubclass(Widget)
class CoactiveWidgetStub(Widget):
    def __init__(self, view, css_id, coactive_widgets):
        super().__init__(view)
        self._coactive_widgets = coactive_widgets
        self.css_id = css_id

    def render_contents(self): 
        return '<%s>' % self.css_id

    @property
    def coactive_widgets(self):
        return super().coactive_widgets + self._coactive_widgets


@with_fixtures(WebFixture)
def test_coactive_widgets(web_fixture):
    """Coactive Widgets of a Widget are Widgets that are included in a WidgetResult for that Widget.

    Included are: the coactive widgets of children of the result widget as well as the coactive widgets of coactive widgets.
    """

    @stubclass(Widget)
    class WidgetWithRemoteMethod(Widget):
        def __init__(self, view):
            super().__init__(view)
            coactive_widgets = [self.add_child(CoactiveWidgetStub(view, 'coactive1', [self.add_child(CoactiveWidgetStub(view, 'coactive2', []))]))]
            result_widget = self.add_child(CoactiveWidgetStub(view, 'main', []))
            result_widget.add_child(CoactiveWidgetStub(view, 'child', coactive_widgets))
            method_result = WidgetResult([result_widget])
            remote_method = RemoteMethod(view, 'amethod', lambda: None, default_result=method_result, disable_csrf_check=True)
            view.add_resource(remote_method)

    wsgi_app = web_fixture.new_wsgi_app(child_factory=WidgetWithRemoteMethod.factory())
    browser = Browser(wsgi_app)

    browser.post('/_amethod_method', {})
    json_response = json.loads(browser.raw_html)
    assert json_response == {'success': True,
                             'exception': '',
                             'result': {
                                'main': '<main><script type="text/javascript"></script>',
                                'coactive1': '<coactive1><script type="text/javascript"></script>',
                                'coactive2': '<coactive2><script type="text/javascript"></script>'}
                             }


@uses(web_fixture=WebFixture)
class CoactiveScenarios(Fixture):
    expected_exception = NoException
    expected_exception_regex = None
    def new_page(self):
        return Div(self.view)

    def new_view(self):
        return self.web_fixture.view

    @scenario
    def descendents_that_are_coactive_are_ignored(self):
        """Descendents that are also coactive_widgets are ignored in the final coactive_widget list."""

        coactive_descendent = CoactiveWidgetStub(self.view, 'coactive_descendent', [])

        widget = self.page.add_child(CoactiveWidgetStub(self.view, 'main', [coactive_descendent]))
        child = widget.add_child(CoactiveWidgetStub(self.view, 'child', []))
        child.add_child(coactive_descendent)

        self.result_widget = widget
        self.expected_coactive_widgets = []
        assert self.result_widget.coactive_widgets == [coactive_descendent]

    @scenario
    def coactive_widgets_include_children_coactive_widgets(self):
        """The coactive widgets of descendents are included in a widget's coactive widgets."""

        coactive_widget = self.page.add_child(CoactiveWidgetStub(self.view, 'coactive_widget', []))
        widget = self.page.add_child(CoactiveWidgetStub(self.view, 'main', []))
        child = widget.add_child(CoactiveWidgetStub(self.view, 'child', [coactive_widget]))

        self.result_widget = widget
        self.expected_coactive_widgets = [coactive_widget]

    @scenario
    def coactiveness_is_transitive(self):
        """The coactive widgets of the coactive widgets of a widget are included in its coactive widgets."""

        coactive_widget_level2 = self.page.add_child(CoactiveWidgetStub(self.view, 'coactive_widget_level2', []))
        coactive_widget_level1 = self.page.add_child(CoactiveWidgetStub(self.view, 'coactive_widget_level1', [coactive_widget_level2]))
        widget = self.page.add_child(CoactiveWidgetStub(self.view, 'main', [coactive_widget_level1]))

        self.result_widget = widget
        self.expected_coactive_widgets = [coactive_widget_level1, coactive_widget_level2]

    @scenario
    def coactive_widgets_cannot_be_parents(self):
        """The ancestor Widgets of a given Widget cannot be its coactive Widgets."""

        grandparent = self.page.add_child(Div(self.view, css_id='grandparent'))
        parent = grandparent.add_child(CoactiveWidgetStub(self.view, 'parent', []))
        child = parent.add_child(CoactiveWidgetStub(self.view, 'child', [grandparent]))

        self.result_widget = child
        self.expected_coactive_widgets = None
        self.expected_exception = ProgrammerError
        self.expected_exception_regex = 'The coactive Widgets of .+ include its ancestor\(s\): .+'


@with_fixtures(WebFixture, CoactiveScenarios)
def test_what_is_included_in_coactive_widgets(web_fixture, coactive_scenarios):

    fixture = coactive_scenarios
    fixture.view.page = fixture.page

    with expected(fixture.expected_exception, fixture.expected_exception_regex):
        coactive_widgets = WidgetResult([fixture.result_widget]).get_coactive_widgets_recursively(fixture.result_widget)
        assert set(coactive_widgets) == set(fixture.expected_coactive_widgets)





# Copyright 2015-2024 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
""".. versionadded:: 3.2

Reahl uses CSS and JavaScript from other projects. The CSS and JavaScript world
has its own ecosystem of tools that allow you to do various things, for example:

 - Compile your own customised CSS using a CSS preprocessor like `Sass <http://sass-lang.com/>`_
 - Manage the versions and interdependencies of such projects.
 - Some code is hosted on Content Delivery Networks (CDNs).
 - Some code is minified.

This module is a start towards supporting such "front-end frameworks"
which Reahl depends upon. By default, Reahl merely includes the correct
versions of the front-end libraries it needs.

Reahl also contains its own JavaScript and CSS code. Where we do this,
our own code is also managed using this same infrastructure.

If you write your own Widgets that make use of CSS or JavaScript,
creating a :class:`Library` and configure your site to use it.

"""

import itertools
from collections import OrderedDict

from reahl.component.context import ExecutionContext
from reahl.component.exceptions import ProgrammerError
from reahl.web.fw import PackagedFile, ConcatenatedFile


class LibraryIndex:
    """An ordered collection of :class:`Library` instances.

    A LibraryIndex instance is available in the configuration as
    `web.config.frontend_libraries`. All the :class:`Library`
    instances contained in `web.config.frontend_libraries` are
    included on any :class:`~reahl.web.ui.HTML5Page`.

    Libraries in a LibraryIndex are included on a page in the order
    that they appear in the LibraryIndex.

    :param libraries: :class:`Library` instances (in order) to initialise this LibraryIndex with.

    """
    def __init__(self, *libraries):
        self.libraries_by_name = OrderedDict()
        for library in libraries:
            self.add(library)

    def clear(self):
        """Removes all Libraries from the list."""
        self.libraries_by_name.clear()

    def add(self, library):
        """Adds a Library to the end of the list."""
        self.libraries_by_name[library.name] = library
        return library

    def get(self, library_class):
        matching_libraries = [i for i in self.libraries_by_name.values() if isinstance(i, library_class)]
        if not matching_libraries:
            raise ProgrammerError('No library "%s" found in the config.web.frontend_libraries' % library_class)
        elif len(matching_libraries) > 1:
            raise ProgrammerError('More than "%s" found in the config.web.frontend_libraries' % library_class)
        return matching_libraries[0]

    def __contains__(self, name):
        """An implementation of the `in` operator, so that one can ask whether a library with given name is in this index.

        :param name: The unique name of the Library to check for.
        """
        return name in self.libraries_by_name

    def packaged_files(self):
        return [i for i in itertools.chain(*[library.packaged_files() for library in self])]

    def __iter__(self):
        return iter(self.libraries_by_name.values())

    def prepend(self, library):
        new_libraries = [library] + list(self.libraries_by_name.values())
        self.libraries_by_name = OrderedDict()
        for library in new_libraries:
            self.add(library)

    def __str__(self):
        return  '%s(%s)' % (self.__class__.__name__, ','.join(self.libraries_by_name.keys()))

class Library:
    """A frontend-library: a collection of CSS and JavaScript code that can be used with Reahl.

    To create your own Library, subclass from this class and set its
    attributes to indicate which files to include as part of it and
    where to find them.

    To use a library, add it to the `web.config.frontend_libraries`
    config setting. The CSS and JavaScript files of all such
    configured libraries are automatically included in any
    :class:`~reahl.web.ui.HTML5Page`.

    :param name: A unique name for this Library.
    """
    active = True
    @classmethod
    def get_instance(cls):
        return ExecutionContext.get_context().config.web.frontend_libraries.get(cls)

    def __init__(self, name):
        self.name = name  #: The unique name of this Library
        self.egg_name = 'reahl-web'  #: The component (egg) that contains the files of this library
        self.shipped_in_package = ''  #: The package that contains the files of this library
        self.files = []   #: The JavaScript and CSS files that form part of this library (relative to the `shipped_in_package`)

    def packaged_files(self):
        return [PackagedFile(self.egg_name, self.shipped_in_package, file_name)
                for file_name in self.files]

    def packaged_files222xxxxxFIXMEIAMDEADButPossibleAlternative(self):
        exposed_files = []
        js_files_to_include = [PackagedFile(self.egg_name, self.shipped_in_package, file_name)
                               for file_name in self.files
                               if file_name.endswith('.js')]
        if js_files_to_include:
            exposed_files.append(ConcatenatedFile('%s.js' % self.name, js_files_to_include))

        css_files_to_include = [PackagedFile(self.egg_name, self.shipped_in_package, file_name)
                               for file_name in self.files
                               if file_name.endswith('.css')]
        if css_files_to_include:
            exposed_files.append(ConcatenatedFile('%s.css' % self.name, css_files_to_include))
        return exposed_files

    def files_of_type(self, extension):
        return [f for f in self.files if f.endswith(extension)]

    def header_only_material(self, rendered_page):
        result = ''
        if self.active:
            for file_name in self.files_of_type('.css'):
                result += '\n<link rel="stylesheet" href="/static/%s" type="text/css">' % file_name
        return result

    def footer_only_material(self, rendered_page):
        result = ''
        if self.active:
            for file_name in self.files_of_type('.js'):
                result += '\n<script type="text/javascript" src="/static/%s"></script>' % file_name
        return result

    def inline_material(self):
        result = ''
        for file_name in self.files_of_type('.js'):
            result += '\n<script type="text/javascript" src="/static/%s"></script>' % file_name
        return result

    
class JQuery(Library):
    """Version 3.7.1 of `JQuery <https://jquery.com>`_.

    This Library also includes a number of plugins we use internally:

    =================== ==================================
     Plugin              Version
    =================== ==================================
     jquery.validate     1.19.5 (a heavily modified version) (https://github.com/jquery-validation/jquery-validation/releases/tag/1.19.5)
     jquery.ba-bbq       1.3pre
     jquery.blockUI      2.70.0 (https://github.com/malsup/blockui/releases)
     jquery.form         4.3.0 (https://github.com/jquery-form/form/releases)
    =================== ==================================
    """
    def __init__(self):
        super().__init__('jquery')
        self.files = ['jquery-3.7.1/jquery-3.7.1.min.js',
                      'jquery-3.7.1/jquery-3.7.1.min.map']
        self.shipped_in_package = 'reahl.web.static'
        for i in ['jquery.validate-1.19.5.modified.js',
                  'jquery.ba-bbq-1.3pre.js',
                  'jquery.blockUI-2.70.0.js',
                  'jquery.form-4.3.0.js']:
            self.add_shipped_plugin('jquery/%s' % i)

    def add_shipped_plugin(self, file_name):
        self.files.append(file_name)
        return file_name

    def document_ready_material(self, rendered_page):
        result = '\n<script id="reahl-jqueryready" type="text/javascript">\n'
        result += 'jQuery(document).ready(function($){\n'
        result += '$(\'body\').addClass(\'enhanced\');\n'
        js = set()
        js.update(rendered_page.get_js())
        for item in sorted(js):
            result += item+'\n'
        result += '\n});'
        result += '\n</script>\n'
        return result

    def footer_only_material(self, rendered_page):
        result = super().footer_only_material(rendered_page)
        #from http://ryanpricemedia.com/2008/03/19/jquery-broken-in-internet-explorer-put-your-documentready-at-the-bottom/
        result += self.document_ready_material(rendered_page)
        return result


class JQueryUI(Library):
    """A heavily customised subset of version 1.13.2 of `JQuery UI <https://jqueryui.com>`_.
    
   Only contains the `Widget Factory <http://api.jqueryui.com/jQuery.widget/>`_ and :tabbable, :focusable Selector.
    """
    def __init__(self):
        super().__init__('jqueryui')
        self.shipped_in_package = 'reahl.web.static'
        self.files = ['jquery-ui-1.13.2.custom/jquery-ui.min.js']


class HTML5Shiv(Library):
    """Version 3.7.3 of `html5shiv <https://github.com/aFarkas/html5shiv>`_.
    
    """
    def __init__(self):
        super().__init__('html5shiv')
        self.shipped_in_package = 'reahl.web.static'
        self.files = ['html5shiv-printshiv-3.7.3.js']

    def footer_only_material(self, rendered_page):
        # From: http://remysharp.com/2009/01/07/html5-enabling-script/ 
        result  = '\n<!--[if lt IE 9]>'
        result  += super().footer_only_material(rendered_page)
        result  += '<![endif]-->'
        return result


class IE9(Library):
    """Version 2.1(beta4) of `IE9.js <https://code.google.com/archive/p/ie7-js/>`_.
    """
    def __init__(self):
        super().__init__('IE9')
        self.shipped_in_package = 'reahl.web.static'
        self.files = ['IE9.js']

    def footer_only_material(self, rendered_page):
        # From: http://code.google.com/p/ie7-js/ 
        # Not sure if this does not perhaps interfere with Normalize reset stuff? 
        result  = '\n<!--[if lte IE 9]>'
        result  += super().footer_only_material(rendered_page)
        result  += '<![endif]-->'
        return result


class Reahl(Library):
    """JavaScript and CSS that is part of Reahl itself.
    """
    def __init__(self):
        super().__init__('reahl')
        self.shipped_in_package = 'reahl.web.static'
        self.files = ['reahl.csrf.js',
                      'reahl.hashchange.js',
                      'reahl.ajaxlink.js',
                      'reahl.primitiveinput.js',
                      'reahl.textinput.js',
                      'reahl.validate.js',
                      'reahl.form.js',
                      'reahl.plotlychart.js',
                      'reahl.css',
                      'reahl.runningonbadge.css',
                      'runningon.png'
                      ]


class Holder(Library):
    """Version 2.9.9 of `Holder <http://imsky.github.io/holder/>`_.
    """
    def __init__(self):
        super().__init__('holder')
        self.shipped_in_package = 'reahl.web.holder'
        self.files = ['holder-2.9.9.js']


class Bootstrap4(Library):
    """Version 4.6.2 of `Bootstrap <http://getbootstrap.com/>`_.
    """
    def __init__(self):
        super().__init__('bootstrap4')
        self.shipped_in_package = 'reahl.web.static'
        self.files = [
                      'bootstrap-4.6.2/css/bootstrap.css',
                      'bootstrap-4.6.2/css/reahl-patch.css',
                      'bootstrap-4.6.2/css/bootstrap.css.map',
                      'bootstrap-4.6.2/js/bootstrap.min.js',
                      'bootstrap-4.6.2/js/bootstrap.min.js.map'
                      ]


    def header_only_material(self, rendered_page):
        return '<meta charset="utf-8">'\
               '<meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">' +\
               super().header_only_material(rendered_page) 


class ReahlBootstrap4Additions(Library):
    """Reahl specific JavaScript and CSS for use with :class:`Bootstrap4`.
    """
    def __init__(self):
        super().__init__('bootstrap4.reahladditions')
        self.shipped_in_package = 'reahl.web.bootstrap'
        self.files = [
                      'reahl.bootstrapform.js',
                      'reahl.pagination.js',
                      'reahl.files.js',
                      'reahl.files.css',
                      'reahl.bootstrappopupa.js',
                      'reahl.carousel.css',
                      'reahl.bootstrapcueinput.js',
                      'reahl.bootstrapfileuploadli.js',
                      'reahl.bootstrapfileuploadpanel.js',
                      'reahl.datatable.css'
                      ]


class Popper(Library):
    """Version 1.16.1 (umd) of `Popper <https://popper.js.org/>`_.
    """
    def __init__(self):
        super().__init__('popper')
        self.shipped_in_package = 'reahl.web.static'
        self.files = [
            'popper-1.16.1/popper.min.js' #make sure it is the umd edition
        ]


class Underscore(Library):
    """Version 1.13.6 of `Underscore.js <https://underscorejs.org>`_.
    """
    def __init__(self):
        super().__init__('underscore')
        self.shipped_in_package = 'reahl.web.static'
        self.files = [
            'underscore-umd-min.1.13.6.js'
        ]
    def footer_only_material(self, rendered_page):
        return super().footer_only_material(rendered_page) + '<script>var underscore = _;</script>'


class JsCookie(Library):
    """Version 3.0.5 of `js-cookie <https://github.com/js-cookie/js-cookie>`_.
    """
    def __init__(self):
        super().__init__('js-cookie')
        self.shipped_in_package = 'reahl.web.static'
        self.files = [
            'js-cookie-3.0.5/js.cookie.min.js' #this is the UMD version
        ]


class PlotlyJS(Library):
    """Version 2.29.0 of `plotly.js <https://github.com/plotly/plotly.js/>`_.
    """
    javascript_filename = 'plotly-2.29.0.min.js'
    def __init__(self):
        self.active = False
        super().__init__('plotly.js')
        self.shipped_in_package = 'reahl.web.static'
        self.files = [
            self.javascript_filename
        ]



# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import os.path

import pkg_resources

from reahl.tofu import scenario, temp_dir, temp_file_with, Fixture
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import easter_egg, stubclass

from reahl.web.fw import FileOnDisk, FileFromBlob, PackagedFile, ConcatenatedFile, FileDownload, UserInterface
from reahl.browsertools.browsertools import Browser

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_files_from_disk(web_fixture):
    """A directory in the web.static_root configuration setting, can be mounted on a URL
       named after it on the WebApplication.
    """

    static_root = temp_dir()
    files_dir = static_root.sub_dir('staticfiles')
    sub_dir = files_dir.sub_dir('subdir')
    one_file = files_dir.file_with('one_file.xml', 'one')
    nested_file = sub_dir.file_with('nested_file', 'other')

    # How the config is set
    web_fixture.config.web.static_root = static_root.name

    # How the subdirectory is mounted
    class MainUI(UserInterface):
        def assemble(self):
            self.define_static_directory('/staticfiles')

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # How the first file would be accessed
    browser.open('/staticfiles/one_file.xml')
    assert browser.raw_html == 'one'

    # The meta-info of the file
    response = browser.last_response
    assert response.content_type == 'application/xml'
    assert response.content_encoding is None
    assert response.content_length == 3
    expected_mtime = datetime.datetime.fromtimestamp(int(os.path.getmtime(one_file.name)))
    assert response.last_modified.replace(tzinfo=None) == expected_mtime
    expected_tag = '%s-%s-%s' % (os.path.getmtime(one_file.name), 3, abs(hash(one_file.name)))
    assert response.etag == expected_tag

    # How the file in the subdirectory would be accessed
    browser.open('/staticfiles/subdir/nested_file')
    assert browser.raw_html == 'other'

    # When a file does not exist
    browser.open('/staticfiles/one_that_does_not_exist', status=404)


@with_fixtures(WebFixture)
def test_files_from_list(web_fixture):
    """An explicit list of files can also be added on an URL as if they were in a single
       directory.
    """

    files_dir = temp_dir()
    one_file = files_dir.file_with('any_name_will_do_here', 'one')

    class MainUI(UserInterface):
        def assemble(self):
            list_of_files = [FileOnDisk(one_file.name, 'one_file')]
            self.define_static_files('/morestaticfiles', list_of_files)

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # How the first file would be accessed
    browser.open('/morestaticfiles/one_file')
    assert browser.raw_html == 'one'

    # The meta-info of the file
    response = browser.last_response
    assert response.content_type == 'application/octet-stream'
    assert response.content_encoding is None
    assert response.content_length == 3
    expected_mtime = datetime.datetime.fromtimestamp(int(os.path.getmtime(one_file.name)))
    assert response.last_modified.replace(tzinfo=None) == expected_mtime
    expected_tag = '%s-%s-%s' % (os.path.getmtime(one_file.name), 3, abs(hash(one_file.name)))
    assert response.etag == expected_tag

    # When a file does not exist
    browser.open('/morestaticfiles/one_that_does_not_exist', status=404)


@with_fixtures(WebFixture)
def test_files_from_database(web_fixture):
    """Files can also be created on the fly such as from data in a database."""


    content_bytes = ('hôt stuff').encode('utf-8')

    class MainUI(UserInterface):
        def assemble(self):
            mime_type = 'text/html'
            encoding = 'utf-8'
            mtime = 123
            size = len(content_bytes)

            list_of_files = [FileFromBlob('database_file', content_bytes, mime_type, encoding, size, mtime)]
            self.define_static_files('/files', list_of_files)

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    # How the file would be accessed
    browser.open('/files/database_file')
    assert browser.raw_html == 'hôt stuff'
    response = browser.last_response

    # The meta-info of the file
    assert response.content_type == 'text/html'
    assert not response.content_encoding
    assert response.content_length == len(content_bytes)
    assert response.last_modified.replace(tzinfo=None) == datetime.datetime.fromtimestamp(123)
    expected_etag = '123-%s-%s' % (len(content_bytes), abs(hash('database_file')))
    assert response.etag == expected_etag


@with_fixtures(WebFixture)
def test_packaged_files(web_fixture):
    """Files can also be served straight from a python egg."""


    # Create an egg with package packaged_files, containing the file packaged_file
    egg_dir = temp_dir()
    package_dir = egg_dir.sub_dir('packaged_files')
    init_file = package_dir.file_with('__init__.py', '')
    afile = package_dir.file_with('packaged_file', 'contents')

    easter_egg.clear()
    pkg_resources.working_set.add(easter_egg)
    easter_egg.location = egg_dir.name

    with easter_egg.active():
        class MainUI(UserInterface):
            def assemble(self):
                list_of_files = [PackagedFile(easter_egg.as_requirement_string(), 'packaged_files', 'packaged_file')]
                self.define_static_files('/files', list_of_files)

        wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the file would be accessed
        browser.open('/files/packaged_file')
        assert browser.raw_html == 'contents'

class ConcatenateScenarios(Fixture):
    @scenario
    def normal_files(self):
        self.file1_contents = 'contents1'
        self.file2_contents = 'contents2'
        self.filename = 'concatenated'
        self.expected_result = 'contents1contents2'

    @scenario
    def javascript_files(self):
        self.file1_contents = 'acall1(); //some comment'
        self.file2_contents = 'acall2(); //some comment'
        self.filename = 'concatenated.js'
        self.expected_result = 'acall1();acall2();'

    @scenario
    def css_files(self):
        self.file1_contents = '.cool {  text:bold;   }/* a comment */ '
        self.file2_contents = 'a, p { text:white;     } '
        self.filename = 'concatenated.css'
        self.expected_result = '.cool{text:bold}a,p{text:white}'


@with_fixtures(WebFixture, ConcatenateScenarios)
def test_concatenated_files(web_fixture, concatenate_scenarios):
    """Files can also be formed by concatenating other files.  Files ending in .css or .js are appropriately
       minified in the process."""

    fixture = concatenate_scenarios

    # Make an egg with a package called packaged_files, and two files in there.
    egg_dir = temp_dir()
    package_dir = egg_dir.sub_dir('packaged_files')
    init_file = package_dir.file_with('__init__.py', '')
    afile = package_dir.file_with('packaged_file', fixture.file1_contents)
    another_file = package_dir.file_with('packaged_file2', fixture.file2_contents)

    pkg_resources.working_set.add(easter_egg)
    easter_egg.location = egg_dir.name

    class MainUI(UserInterface):
        def assemble(self):
            to_concatenate = [PackagedFile('test==1.0', 'packaged_files', 'packaged_file'),
                              PackagedFile('test==1.0', 'packaged_files', 'packaged_file2')]
            list_of_files = [ConcatenatedFile(fixture.filename, to_concatenate)]
            self.define_static_files('/files', list_of_files)

    with easter_egg.active():
        web_fixture.config.reahlsystem.debug = False  # To enable minification
        wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        # How the first file would be accessed
        browser.open('/files/%s' % fixture.filename)
        assert browser.raw_html == fixture.expected_result


@with_fixtures(WebFixture)
def test_file_download_details(web_fixture):
    """FileDownloadStub (the GET response for a StaticFileResource) works correctly in
      different scenarios of partial GETting too."""


    file_content = b'some content'
    server_file = temp_file_with(file_content, 'afile.css', mode='w+b')
    @stubclass(FileDownload)
    class FileDownloadStub(FileDownload):
        chunk_size = 1
    response = FileDownloadStub(FileOnDisk(server_file.name, '/path/for/the/file'))

    # Case: The whole content is sent, in chunk_size bits
    read = [i for i in response.app_iter]
    expected = [bytes((i,)) for i in file_content]
    assert read == expected

    # Case: Headers are set correctly
    assert response.content_type == 'text/css'
    assert not response.content_encoding
    assert response.content_length == len(file_content)

    mtime = datetime.datetime.fromtimestamp(int(os.path.getmtime(server_file.name)))
    assert response.last_modified.replace(tzinfo=None) == mtime
    tag_mtime, tag_size, tag_hash = response.etag.split('-')
    mtime = str(os.path.getmtime(server_file.name))
    assert tag_mtime == mtime
    assert tag_size == str(len(file_content))
    assert tag_hash == str(abs(hash(server_file.name)))

    # Case: conditional response is supported
    assert response.conditional_response

    # Case: partial response is supported - different cases:
    #      - normal case
    actual = [i for i in response.app_iter.app_iter_range(3,7)]
    expected = [bytes((i,)) for i in file_content[3:8]]
    assert actual == expected

    #      - no end specified
    actual = [i for i in response.app_iter.app_iter_range(3)]
    expected = [bytes((i,)) for i in file_content[3:]]
    assert actual == expected

    #      - no start specified
    actual = [i for i in response.app_iter.app_iter_range(end=7)]
    expected = [bytes((i,)) for i in file_content[:8]]
    assert actual == expected

    #      - where the last chunk read would stretch past end
    response.chunk_size = 2
    actual = b''.join([i for i in response.app_iter.app_iter_range(end=6)])
    expected = file_content[:7]
    assert actual == expected

    #      - where start > end
    response.chunk_size = 1
    actual = [i for i in response.app_iter.app_iter_range(start=7, end=3)]
    expected = [b'']
    assert actual == expected

    #      - where start < 0
    actual = [i for i in response.app_iter.app_iter_range(start=-10, end=7)]
    expected = [bytes((i,)) for i in file_content[:8]]
    assert actual == expected

    #      - where end > length of file
    actual = [i for i in response.app_iter.app_iter_range(start=3, end=2000)]
    expected = [bytes((i,)) for i in file_content[3:]]
    assert actual == expected

    #      - where start > length of file
    actual = [i for i in response.app_iter.app_iter_range(start=700)]
    expected = [b'']
    assert actual == expected



# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import os.path
import warnings
import io
import itertools

from reahl.tofu import expected, scenario, Fixture, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import EmptyStub

from reahl.browsertools.browsertools import Browser, XPath

from reahl.component.exceptions import ProgrammerError, IncorrectArgumentError, IsSubclass
from reahl.web.fw import UserInterface
from reahl.web.ui import HTML5Page, P, Div

from reahl.web_dev.fixtures import WebFixture, BasicPageLayout


@uses(web_fixture=WebFixture)
class BasicScenarios(Fixture):

    def get_file_content(self, filename):
        with(io.open('%s/%s' % (os.path.dirname(__file__), filename))) as f:
            file_content = ''.join(f.readlines())
        return file_content[:-1]

    @property
    def view(self):
        return self.web_fixture.view

    expected_warnings = []
    @scenario
    def view_with_page(self):
        class SimplePage(HTML5Page):
            def __init__(self, view):
                super().__init__(view)
                self.body.add_child(P(view, text='Hello world!'))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_view('/', title='Hello', page=SimplePage.factory())

        self.MainUI = MainUI
        self.content_includes_p = True

    @scenario
    def view_with_set_page(self):
        class SimplePage(HTML5Page):
            def __init__(self, view):
                super().__init__(view)
                self.body.add_child(P(view, text='Hello world!'))

        class MainUI(UserInterface):
            def assemble(self):
                home = self.define_view('/', title='Hello')
                home.set_page(SimplePage.factory())

        self.MainUI = MainUI
        self.content_includes_p = True

    @scenario
    def ui_with_page(self):
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_view('/', title='Hello')

        self.MainUI = MainUI
        self.content_includes_p = False


@with_fixtures(WebFixture, BasicScenarios)
def test_basic_assembly(web_fixture, basic_scenarios):
    """An application is built by extending UserInterface, and defining this UserInterface in an .assemble() method.

    To define the UserInterface, several Views are defined. Views are mapped to URLs. When a user GETs
    the URL of a View, a page is rendered back to the user. How that page is created
    can happen in different ways, as illustrated by each scenario of this test.
    """
    fixture = basic_scenarios

    wsgi_app = web_fixture.new_wsgi_app(site_root=fixture.MainUI)
    browser = Browser(wsgi_app)

    # GETting the URL results in the HTML for that View
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')
        browser.open('/')
        assert browser.title == 'Hello'

    warning_messages = [str(i.message) for i in caught_warnings]
    assert len(warning_messages) == len(fixture.expected_warnings)
    for caught, expected_message in itertools.zip_longest(warning_messages, fixture.expected_warnings):
        assert expected_message in caught

    if fixture.content_includes_p:
        [message] = browser.xpath('//p')
        assert message.text == 'Hello world!'

    # The headers are set correctly
    response = browser.last_response
    assert response.content_type == 'text/html'
    assert response.charset == 'utf-8'
    assert str(response.cache_control) == 'no-store'

    # Invalid URLs do not exist
    with warnings.catch_warnings(record=True):
        browser.open('/nonexistantview/', status=404)


@with_fixtures(WebFixture)
def test_basic_error1(web_fixture):
    """Sending the the wrong kind of thing as widget_class to define_page is reported to the programmer."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(EmptyStub)
            self.define_view('/', title='Hello')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(IsSubclass):
        browser.open('/')


@with_fixtures(WebFixture)
def test_basic_error2(web_fixture):
    """Sending the the wrong arguments for the specified class to define_page is reported to the programmer."""

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page, 1, 2)
            self.define_view('/', title='Hello')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(IncorrectArgumentError, test='define_page was called with arguments that do not match those expected by.*'):
        browser.open('/')


@with_fixtures(WebFixture)
def test_basic_error3(web_fixture):
    """Forgetting to define either a page of a page for a View is reported to the programmer."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_view('/', title='Hello')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(ProgrammerError, test='there is no page defined for /'):
        browser.open('/')


@uses(web_fixture=WebFixture)
class SlotScenarios(Fixture):
    @scenario
    def page_on_ui(self):
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(BasicPageLayout())
                home = self.define_view('/', title='Hello')
                home.set_slot('main', P.factory(text='Hello world'))
                home.set_slot('footer', P.factory(text='I am the footer'))
        self.MainUI = MainUI

    @scenario
    def page_on_view(self):
        class MainUI(UserInterface):
            def assemble(self):
                home = self.define_view('/', title='Hello')
                home.set_page(HTML5Page.factory().use_layout(BasicPageLayout()))
                home.set_slot('main', P.factory(text='Hello world'))
                home.set_slot('footer', P.factory(text='I am the footer'))
        self.MainUI = MainUI


@with_fixtures(WebFixture, SlotScenarios)
def test_slots(web_fixture, slot_scenarios):
    """A View modifies the page by populating named Slots in the page with Widgets."""
    fixture = slot_scenarios

    wsgi_app = web_fixture.new_wsgi_app(site_root=fixture.MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')
    assert browser.title == 'Hello'
    [main_p, footer_p] = browser.xpath('//p')
    assert main_p.text == 'Hello world'
    assert footer_p.text == 'I am the footer'


@with_fixtures(WebFixture)
def test_slot_error(web_fixture):
    """Supplying contents for a slot that does not exist results in s sensible error."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Hello')
            home.set_slot('main', P.factory(text='Hello world'))
            home.set_slot('nonexistantslotname', P.factory(text='I am breaking'))

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(ProgrammerError, test='An attempt was made to plug Widgets into the following slots that do not exist.*'):
        browser.open('/')


@with_fixtures(WebFixture)
def test_slot_defaults(web_fixture):
    """If a View does not specify contents for a Slot, the Slot will be populated by the window's default
       widget for that slot if specified, else it will be left empty.
    """
    class MainUI(UserInterface):
        def assemble(self):
            main = self.define_page(HTML5Page).use_layout(BasicPageLayout())
            main.add_default_slot('main', P.factory(text='defaulted slot contents'))
            self.define_view('/', title='Hello')

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')

    # The default widget for the main slot is used
    [p] = browser.xpath('//p')
    assert p.text == 'defaulted slot contents'

    # The header slot has no default, and is thus left empty
    header_contents = browser.xpath('//header/*')
    assert not header_contents


@with_fixtures(WebFixture)
def test_out_of_bound_widgets(web_fixture):
    """When you need to add a widget to the page, but not as a child/descendant."""

    class MyPanel(Div):
        def __init__(self, view):
            super().__init__(view, css_id='main_panel')
            child_widget = self.add_child(P(view, text='Child Widget'))
            out_of_bound_widget = view.add_out_of_bound_widget(P(view, text='Out Of Bound Widget'))

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            home = self.define_view('/', title='Hello')
            home.set_slot('main', MyPanel.factory())

    fixture = web_fixture

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')

    main_panel = XPath.div().with_id('main_panel')
    assert browser.is_element_present(XPath.paragraph().with_text('Out Of Bound Widget'))
    assert not browser.is_element_present(XPath.paragraph().with_text('Out Of Bound Widget').inside_of(main_panel))
    assert browser.is_element_present(XPath.paragraph().with_text('Child Widget').inside_of(main_panel))



# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.stubble import stubclass
from reahl.tofu import Fixture, scenario, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import XPath
from reahl.web.bootstrap.ui import P, Div
from reahl.web.bootstrap.pagination import PageMenu, PagedPanel, PageIndex, SequentialPageIndex, AnnualItemOrganiserProtocol, AnnualPageIndex

from reahl.web_dev.fixtures import WebFixture

@uses(web_fixture=WebFixture)
class PageMenuFixture(Fixture):

    def new_number_of_pages(self):
        return 10

    def new_max_page_links(self):
        return 5

    def new_PageIndexStub(self):
        fixture = self
        @stubclass(PageIndex)
        class PageIndexStub(PageIndex):
            def __init__(self, max_page_links, number_of_pages):
                super().__init__(max_page_links=max_page_links)
                self.number_of_pages = number_of_pages

            @property
            def total_number_of_pages(self):
                return self.number_of_pages

            def get_contents_for_page(self, page_number):
                return 'contents of page %s' % page_number

            def get_description_for_page(self, page_number):
                return 'p%s' % page_number

        return PageIndexStub

    def new_PageContainer(self):
        class PageContainer(PagedPanel):
            def __init__(self, parent, page_index):
                super().__init__(parent, page_index, 'container')
                self.add_child(P(self.view, text=self.current_contents))

        return PageContainer

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Div):
            def __init__(self, view):
                super().__init__(view)
                page_index = fixture.PageIndexStub(fixture.max_page_links, fixture.number_of_pages)
                page_container = self.add_child(fixture.PageContainer(self.view, page_index))
                self.add_child(PageMenu(self.view, 'page_menu_widget', page_index, page_container))
        return MainWidget

    def container_contents_is(self, expected):
        return self.web_fixture.driver_browser.execute_script('return window.jQuery("div p").html() == "%s"' % expected)

    def page_range_links_match(self, link_labels):
        return self.web_fixture.driver_browser.execute_script('return window.jQuery(".pagination a").slice(2,-2).map(function(){return window.jQuery(this).html();}).toArray() == "%s"' % link_labels)

    def is_marked_active(self, link_label, nth=1):
        [li] = self.web_fixture.driver_browser.xpath('%s/..' % (XPath.link().with_text(link_label).inside_of(XPath.ul().including_class('reahl-menu')[nth])))
        return 'active' in li.attrib.get('class', '')

    def new_wsgi_app(self):
        return self.web_fixture.new_wsgi_app(enable_js=True, child_factory=self.MainWidget.factory())


# Please see fixture for how to declare a PageIndex, PageContainer and PageMenu
@with_fixtures(WebFixture, PageMenuFixture)
def test_selecting_a_page(web_fixture, page_menu_fixture):
    """Clicking the link of a page results in the contents of the PageContainer being refreshed."""

    web_fixture.reahl_server.set_app(page_menu_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.wait_for(page_menu_fixture.container_contents_is, 'contents of page 1')
    browser.click(XPath.link().with_text('p2'))
    browser.wait_for(page_menu_fixture.container_contents_is, 'contents of page 2')


@with_fixtures(WebFixture, PageMenuFixture)
def test_navigating_the_page_numbers(web_fixture, page_menu_fixture):
    """One can navigate the range of page links displayed by the PageMenu using the special links."""

    fixture = page_menu_fixture

    fixture.number_of_pages = 30
    fixture.max_page_links = 5
    web_fixture.reahl_server.set_app(page_menu_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    # Case: next link
    browser.click(XPath.link().with_text_starting('»'))
    assert browser.wait_for(fixture.page_range_links_match, 'p6,p7,p8,p9,p10')
    browser.click(XPath.link().with_text_starting('»'))
    assert browser.wait_for(fixture.page_range_links_match, 'p11,p12,p13,p14,p15')

    # Case: prev link
    browser.click(XPath.link().with_text_starting('«'))
    assert browser.wait_for(fixture.page_range_links_match, 'p6,p7,p8,p9,p10')

    # Case: last link
    browser.click(XPath.link().with_text_starting('→'))
    assert browser.wait_for(fixture.page_range_links_match, 'p26,p27,p28,p29,p30')

    # Case: first link
    browser.click(XPath.link().with_text_starting('←'))
    assert browser.wait_for(fixture.page_range_links_match, 'p1,p2,p3,p4,p5')


@with_fixtures(WebFixture, PageMenuFixture)
def test_contents_when_navigating_the_page_numbers(web_fixture, page_menu_fixture):
    """When navigating the range of page links, the currently displayed contents stay unchanged."""

    page_menu_fixture.number_of_pages = 30
    page_menu_fixture.max_page_links = 5
    web_fixture.reahl_server.set_app(page_menu_fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    browser.click(XPath.link().with_text('p2'))
    browser.wait_for(page_menu_fixture.container_contents_is, 'contents of page 2')
    browser.click(XPath.link().with_text_starting('»'))
    browser.wait_for(page_menu_fixture.container_contents_is, 'contents of page 2')


@with_fixtures(WebFixture, PageMenuFixture)
def test_active_state_of_page_links(web_fixture, page_menu_fixture):
    """When choosing a page, the new page link is marked as active, without a server round-trip."""
    fixture = page_menu_fixture


    fixture.number_of_pages = 30
    fixture.max_page_links = 5
    web_fixture.reahl_server.set_app(fixture.wsgi_app)
    web_fixture.driver_browser.open('/')

    with web_fixture.driver_browser.no_load_expected_for('.pagination>*'):
        assert not fixture.is_marked_active('p2')
        web_fixture.driver_browser.click(XPath.link().with_text('p2'))
        web_fixture.driver_browser.wait_for(fixture.is_marked_active, 'p2')


@with_fixtures(WebFixture, PageMenuFixture)
def test_active_state_on_multiple_menus(web_fixture, page_menu_fixture):
    """If there's more than one PageMenu on the page, the active page is switched for both of them"""
    fixture = page_menu_fixture


    class MainWidget(Div):
        def __init__(self, view):
            super().__init__(view)
            page_index = fixture.PageIndexStub(fixture.max_page_links, fixture.number_of_pages)
            page_container = self.add_child(fixture.PageContainer(self.view, page_index))
            self.add_child(PageMenu(self.view, 'page_menu_widget', page_index, page_container))
            self.add_child(PageMenu(self.view, 'page_menu_widget2', page_index, page_container))
    fixture.MainWidget = MainWidget

    web_fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    assert not fixture.is_marked_active('p2', nth=1)
    assert not fixture.is_marked_active('p2', nth=2)
    browser.click(XPath.link().with_text('p2'))
    browser.wait_for(fixture.is_marked_active, 'p2', 1)
    browser.wait_for(fixture.is_marked_active, 'p2', 2)


@with_fixtures(WebFixture, PageMenuFixture)
def test_active_state_of_next_prev_links(web_fixture, page_menu_fixture):
    """Next and Last links are only active when not on the last range of pages,
       and Prev and First are similarly deactive when on the first range of pages."""
    fixture = page_menu_fixture

    fixture.number_of_pages = 15
    fixture.max_page_links = 5
    web_fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    # Case: when you are on the left of the page range
    assert not browser.is_active(XPath.link().with_text_starting('←'))
    assert not browser.is_active(XPath.link().with_text_starting('«'))
    assert browser.is_active(XPath.link().with_text_starting('»'))
    assert browser.is_active(XPath.link().with_text_starting('→'))

    # Case: when you are in the middle of the page range
    browser.click(XPath.link().with_text_starting('»'))
    browser.wait_for_element_present(XPath.link().with_text('p6'))
    assert browser.is_active(XPath.link().with_text_starting('←'))
    assert browser.is_active(XPath.link().with_text_starting('«'))
    assert browser.is_active(XPath.link().with_text_starting('»'))
    assert browser.is_active(XPath.link().with_text_starting('→'))

    # Case: when you are at the end of the page range
    browser.click(XPath.link().with_text_starting('»'))
    browser.wait_for_element_present(XPath.link().with_text('p11'))
    assert browser.is_active(XPath.link().with_text_starting('←'))
    assert browser.is_active(XPath.link().with_text_starting('«'))
    assert not browser.is_active(XPath.link().with_text_starting('»'))
    assert not browser.is_active(XPath.link().with_text_starting('→'))


class LinkScenarios(PageMenuFixture):
    @scenario
    def only_one_page(self):
        self.number_of_pages = 1
        self.max_page_links = 5
        self.goto_last_range = False
        self.visible_page_descriptions = 'p1'
        self.visible_last_page_descriptions = self.visible_page_descriptions

    @scenario
    def fewer_pages_than_max_links(self):
        self.number_of_pages = 2
        self.max_page_links = 5
        self.goto_last_range = False
        self.visible_page_descriptions = 'p1,p2'
        self.visible_last_page_descriptions = self.visible_page_descriptions

    @scenario
    def more_pages_than_max_links(self):
        self.number_of_pages = 10
        self.max_page_links = 5
        self.goto_last_range = True
        self.visible_page_descriptions = 'p1,p2,p3,p4,p5'
        self.visible_last_page_descriptions = 'p6,p7,p8,p9,p10'

    @scenario
    def more_than_one_range_last_range_not_complete(self):
        self.number_of_pages = 7
        self.max_page_links = 5
        self.goto_last_range = True
        self.visible_page_descriptions = 'p1,p2,p3,p4,p5'
        self.visible_last_page_descriptions = 'p3,p4,p5,p6,p7'

    @scenario
    def changing_number_of_links(self):
        self.number_of_pages = 10
        self.max_page_links = 3
        self.goto_last_range = False
        self.visible_page_descriptions = 'p1,p2,p3'
        self.visible_last_page_descriptions = self.visible_page_descriptions


@with_fixtures(WebFixture, LinkScenarios)
def test_which_links_display(web_fixture, link_scenarios):
    """The menu displays the correct range of page links, depending on the starting page in the range, the
       total number of pages and the max number of links in a range"""
    fixture = link_scenarios
    web_fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = web_fixture.driver_browser
    browser.open('/')
    assert browser.wait_for(fixture.page_range_links_match, fixture.visible_page_descriptions)

    if fixture.goto_last_range:
        browser.click(XPath.link().with_text_starting('→'))

    assert browser.wait_for(fixture.page_range_links_match, fixture.visible_last_page_descriptions)


def test_annual_page_index():
    """The AnnualPageIndex breaks a query of items up into pages containing items with the same year."""

    @stubclass(AnnualItemOrganiserProtocol)
    class ItemOrganiser(AnnualItemOrganiserProtocol):
        def get_years(self):                return [2000, 2001]
        def get_items_for_year(self, year): return [year]*3

    page_index = AnnualPageIndex(ItemOrganiser())

    assert page_index.get_contents_for_page(2) == [2001, 2001, 2001]
    assert page_index.get_description_for_page(1) == '2000'
    assert page_index.years == [2000, 2001]
    assert page_index.total_number_of_pages == 2


class SequentialScenarios(Fixture):
    def new_items(self):
        return list(range(0, self.number_of_items))

    def new_page_index(self):
        return SequentialPageIndex(self.items, items_per_page=self.items_per_page, max_page_links=12)

    @scenario
    def less_than_a_page_items(self):
        self.number_of_items = 2
        self.items_per_page = 10
        self.expected_pages = 1
        self.last_page_contents = self.items[0:2]

    @scenario
    def exactly_one_page_items(self):
        self.number_of_items = 10
        self.items_per_page = 10
        self.expected_pages = 1
        self.last_page_contents = self.items[0:10]

    @scenario
    def just_more_than_one_page_items(self):
        self.number_of_items = 15
        self.items_per_page = 10
        self.expected_pages = 2
        self.last_page_contents = self.items[10:15]


@with_fixtures(SequentialScenarios)
def test_sequential_page_index(sequential_scenarios):
    """The SequentialPageIndex breaks a query of items up into pages of sequential items"""

    fixture = sequential_scenarios

    assert fixture.page_index.total_number_of_pages == fixture.expected_pages
    page = fixture.page_index.get_page_number(fixture.expected_pages)
    assert page.description == str(fixture.expected_pages)
    assert page.contents == fixture.last_page_contents
    assert fixture.page_index.max_page_links == 12


@with_fixtures(WebFixture, PageMenuFixture)
def test_pagination_basics(web_fixture, page_menu_fixture):

    #This is what the pager looks like: ← « 1 2 3 » →
    #With this query string we are indicating that we have clicked on page 3, and thus should be the current page
    web_fixture.request.query_string = 'current_page_number=3&start_page_number=1'
    page_index = page_menu_fixture.PageIndexStub(3, 9)
    page_container = page_menu_fixture.PageContainer(web_fixture.view, page_index)
    pagemenu = PageMenu(web_fixture.view, 'my_page_menu_widget', page_index, page_container)

    [pagination_container] = pagemenu.children
    assert 'pagination' in pagination_container.html_representation.get_attribute('class')

    [items_container] = pagination_container.children
    page_items = items_container.children
    assert len(page_items) == 7

    disabled_page_link_indexes = [0, 1] # These are disabled(unclickable): ← «
    current_page_link_index = 4 # given the query string, the link to page 3, is active
    for i, page_item in enumerate(page_items):
        assert 'page-item' in page_item.get_attribute('class')

        [page_link] = page_item.children
        assert 'page-link' in page_link.get_attribute('class')
        assert 'reahl-ajaxlink' in page_link.get_attribute('class')

        # links you cannot click on (disabled)
        if i in disabled_page_link_indexes:
            assert 'disabled' in page_item.get_attribute('class')

            assert '-1' == page_link.get_attribute('tabindex')
            assert page_link.disabled
        else:
            assert 'disabled' not in page_item.get_attribute('class')

            assert not page_link.has_attribute('tabindex')
            assert not page_link.disabled

        # current page is highlighted if you selected the page(are on the page)
        if i == current_page_link_index:
            assert 'active' in page_item.get_attribute('class')
        else:
            assert 'active' not in page_item.get_attribute('class')



#  Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
# -*- coding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.tofu import Fixture, scenario, expected, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.stubble import EmptyStub

from reahl.component.modelinterface import Field
from reahl.web.fw import Factory, FactoryFromUrlRegex, UserInterface, RegexPath, NoMatchingFactoryFound, UrlBoundView

from reahl.web_dev.fixtures import WebFixture


def test_factory_basics():
    """A Factory can be used to represent a set of parameterised instances instead of having to construct
       or retrieve all instances.  When a specific instance is needed, the Factory creates (or retrieves)
       the instance, based on parameters passed."""

    def create_method(arg, kwarg=None):
        instance = EmptyStub()
        instance.arg = arg
        instance.kwarg = kwarg
        return instance

    factory = Factory(create_method)
    instance = factory.create(1, kwarg=2)
    assert instance.arg == 1
    assert instance.kwarg == 2


def test_factory_failure_to_create():
    """Parameters sent to .create should identify an instance uniquely and so enable the create_method
       to construct the correct instance on request. If the create_method cannot create or retrieve
       an instance using those parameters, it raises NoMatchingFactoryFound to indicate that the asked-for instance
       does not exist."""

    def create_method(arg, kwarg=None):
        raise NoMatchingFactoryFound()

    factory = Factory(create_method)
    with expected(NoMatchingFactoryFound):
        factory.create(1, kwarg=2)


@with_fixtures(WebFixture)
def test_factory_from_path_regex(web_fixture):
    """An FactoryFromUrlRegex parses args from the given URL and passes these as kwargs to create_method
       along with the args and kwargs passed to its .create()."""

    def create_method(path, path_arg=None, extra_kwarg=None):
        instance = EmptyStub()
        instance.path_arg = path_arg
        instance.extra_kwarg = extra_kwarg
        return instance


    argument_fields = {'path_arg': Field()}
    factory = FactoryFromUrlRegex(RegexPath('some(?P<path_arg>.+)path', 'some${path_arg}path', argument_fields),
                                  create_method, dict(extra_kwarg='42'))
    instance = factory.create('somecoolpath')
    assert instance.path_arg == 'cool'
    assert instance.extra_kwarg == '42'


@uses(web_fixture=WebFixture)
class MatchingScenarios(Fixture):

    class ViewWithArg(UrlBoundView):
        def assemble(self, my_one_arg=None): pass

    @scenario
    def slash(self):
        self.matched_path = '/editions'
        self.factory = self.web_fixture.user_interface.define_view('/', title='the view')
        self.is_applicable = False

    @scenario
    def slash_with_subresource(self):
        self.matched_path = '/__a_sub_resource_url'
        self.factory = self.web_fixture.user_interface.define_view('/', title='the view')
        self.is_applicable = True

    @scenario
    def shorter_match_than_discriminator(self):
        self.matched_path = '/editions'
        self.factory = self.web_fixture.user_interface.define_view('/edit', title='the view')
        self.is_applicable = False

    @scenario
    def dynamic_slash_with_args(self):
        self.matched_path = '/editions'
        self.factory = self.web_fixture.user_interface.define_view('/', view_class=self.ViewWithArg, my_one_arg=Field())
        self.is_applicable = True

    @scenario
    def dynamic_slash_with_args_not_enough(self):
        self.matched_path = '/editions/'
        self.factory = self.web_fixture.user_interface.define_view('/', view_class=self.ViewWithArg, my_one_arg=Field())
        self.is_applicable = False

    class UIWithoutKwarg(UserInterface):
        pass

    @scenario
    def parameterised_user_interface_with_long_relative_path(self):
        self.matched_path = '/editions/relative_path'
        self.factory = self.web_fixture.user_interface.define_user_interface('/editions', self.UIWithoutKwarg, {})
        self.is_applicable = True

    @scenario
    def parameterised_user_interface_with_short_relative_path(self):
        self.matched_path = '/editions/'
        self.factory = self.web_fixture.user_interface.define_user_interface('/editions', self.UIWithoutKwarg, {})
        self.is_applicable = True

    @scenario
    def parameterised_user_interface_without_relative_path(self):
        """A user_interface matches an url even if the url does not contain a relative path."""
        self.matched_path = '/editions'
        self.factory = self.web_fixture.user_interface.define_user_interface('/editions', self.UIWithoutKwarg, {})
        self.is_applicable = True

    class UIWithKwarg(UserInterface):
        def assemble(self, my_one_arg=None):
            pass

    @scenario
    def parameterised_user_interface_with_args_and_path(self):
        self.matched_path = '/editions/argument1/'
        self.factory = self.web_fixture.user_interface.define_user_interface('/editions', self.UIWithKwarg, {}, my_one_arg=Field())
        self.is_applicable = True

    @scenario
    def parameterised_user_interface_with_subresources(self):
        self.matched_path = '/editions/argument1/__a_sub_resource'
        self.factory = self.web_fixture.user_interface.define_user_interface('/editions', self.UIWithKwarg, {}, my_one_arg=Field())
        self.is_applicable = True


@with_fixtures(WebFixture, MatchingScenarios)
def test_matching(web_fixture, matching_scenarios):

    applicable_rating = matching_scenarios.factory.is_applicable_for(matching_scenarios.matched_path)
    assert (applicable_rating > 0) == matching_scenarios.is_applicable



import os
from reahl.sqlalchemysupport import SqlAlchemyControl

reahlsystem.root_egg = 'reahl-web'
#reahlsystem.connection_uri = 'postgresql:///reahl'
#reahlsystem.connection_uri = 'sqlite:////tmp/test.db'
#reahlsystem.connection_uri = 'sqlite:///:memory:'
reahlsystem.connection_uri = os.environ.get('REAHL_TEST_CONNECTION_URI', 'postgresql:///reahl')

reahlsystem.orm_control = SqlAlchemyControl(echo=False)







# Copyright 2013-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.




from reahl.tofu import scenario, Fixture, expected, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import EmptyStub, stubclass

from reahl.web.ui import *
from reahl.browsertools.browsertools import WidgetTester

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_basic_fixed_attributes(web_fixture):
    """How the static attributes of a Widget can be manipulated, queried and rendered."""

    fixture = web_fixture

    widget = HTMLElement(fixture.view, 'x')
    tester = WidgetTester(widget)

    # Adding / setting
    widget.set_attribute('attr1', 'value1')
    widget.add_to_attribute('listattr', ['one', 'two'])
    widget.add_to_attribute('listattr', ['three'])

    # Querying
    assert widget.has_attribute('attr1')
    assert widget.has_attribute('listattr')
    assert not widget.has_attribute('notthere')
    assert widget.attributes.v['attr1'] == 'value1'
    assert widget.attributes.v['listattr'] == 'one three two'

    # Rendering
    rendered = tester.render_html()
    assert rendered == '<x attr1="value1" listattr="one three two">'

    # Rendering - the order of attributes in the rendered output
    widget.set_attribute('id', '123')
    widget.add_to_attribute('class', ['z', 'b'])
    rendered = tester.render_html()
    assert rendered == '<x id="123" attr1="value1" listattr="one three two" class="b z">'


@with_fixtures(WebFixture)
def test_handy_methods(web_fixture):
    """Some handy methods for special attributes."""

    fixture = web_fixture

    widget = HTMLElement(fixture.view, 'x')
    tester = WidgetTester(widget)

    widget.set_title('the title')
    widget.set_id('the_id')
    widget.append_class('two')
    widget.append_class('one')

    rendered = tester.render_html()
    assert rendered == '<x id="the_id" title="the title" class="one two">'


@with_fixtures(WebFixture)
def test_dynamically_determining_attributes(web_fixture):
    """A Widget can determine its attribute values at the latest possible stage, based on changing data."""

    class WidgetWithDynamicAttributes(HTMLElement):
        state = '1'
        @property
        def attributes(self):
            attributes = super().attributes
            attributes.set_to('dynamic', self.state)
            attributes.add_to('dynamiclist', [self.state])
            attributes.add_to('not-there', ['a value'])
            attributes.remove_from('not-there', ['a value'])
            return attributes

    fixture = web_fixture

    widget = WidgetWithDynamicAttributes(fixture.view, 'x')
    widget.set_attribute('fixed', 'value1')
    tester = WidgetTester(widget)

    widget.state = '1'
    rendered = tester.render_html()
    assert rendered == '<x dynamic="1" dynamiclist="1" fixed="value1">'

    widget.state = '2'
    rendered = tester.render_html()
    assert rendered == '<x dynamic="2" dynamiclist="2" fixed="value1">'


@with_fixtures(WebFixture)
def test_delegating_setting_of_attributes(web_fixture):
    """One or more DelegatedAttributes instances can be added to an HTMLElement as `attribute_source` -- an
       object which the HTMLElement can use in order to let some other object modify or add to its attributes.
    """

    @stubclass(DelegatedAttributes)
    class MyDelegatedAttributesClass(DelegatedAttributes):
        def set_attributes(self, attributes):
            attributes.set_to('set-by-external-source', 'rhythm and poetry')

    fixture = web_fixture

    widget = HTMLElement(fixture.view, 'x')
    widget.add_attribute_source(MyDelegatedAttributesClass())
    tester = WidgetTester(widget)

    # Case: dynamic attributes are supplied by the wrapper
    rendered = tester.render_html()
    assert rendered == '<x set-by-external-source="rhythm and poetry">'


@with_fixtures(WebFixture)
def test_all_html_widgets_have_css_ids(web_fixture):
    """A Widget (for HTML) can have a css_id. If accessed, but not set, a ProgrammerError is raised."""

    fixture = web_fixture

    widget = HTMLElement(fixture.view, 'x')
    widget.tag_name = 'x'
    tester = WidgetTester(widget)

    widget._css_id = None
    with expected(ProgrammerError):
        widget.css_id

    widget._css_id = 'myid'
    assert widget.css_id == 'myid'


@with_fixtures(WebFixture)
def test_jquery_support(web_fixture):
    """Each HTML Widget has a jquery_selector that targets it uniquely. By default
       this makes use of the css_id of the Widget, which thus need to be set, however
       subclasses may override this behaviour. The jquery_selector (or a more lax
       selector for the Widget) can also be further narrowed to a specific
       jquery selector context."""

    fixture = web_fixture

    widget = HTMLElement(fixture.view, 'x')
    tester = WidgetTester(widget)

    with expected(ProgrammerError):  # css id not set
        widget.jquery_selector

    widget.set_id('anid')
    assert widget.jquery_selector == '"#%s"' % widget.css_id
    assert widget.attributes['id'].as_html_value() == widget.css_id

    contextualised = widget.contextualise_selector('selector', 'context')
    assert contextualised ==  'selector, "context"'

    contextualised = widget.contextualise_selector('selector', None)
    assert contextualised ==  'selector'


@with_fixtures(WebFixture)
def test_single_tags(web_fixture):
    """Definition of a HTMLElement."""

    fixture = web_fixture

    single_tag = HTMLElement(fixture.view, 'single')
    tester = WidgetTester(single_tag)

    with expected(AssertionError):
        single_tag.add_child(P(fixture.view))

    rendered = tester.render_html()
    assert rendered == '<single>'


@with_fixtures(WebFixture)
def test_closing_tags(web_fixture):
    """Definition of a HTMLElement with children."""

    fixture = web_fixture

    closing_tag = HTMLElement(fixture.view, 'closing', children_allowed=True)
    tester = WidgetTester(closing_tag)

    closing_tag.add_child(P(fixture.view))

    rendered = tester.render_html()
    assert rendered == '<closing><p></p></closing>'


@uses(web_fixture=WebFixture)
class WidgetScenarios(Fixture):

    @property
    def view(self):
        return self.web_fixture.view

    @scenario
    def text_node(self):
        self.widget = TextNode(self.view, 'text')
        self.expected_html = 'text'

    @scenario
    def a1(self):
        self.widget = A(self.view, Url('/xyz'))
        self.expected_html = '<a href="/xyz"></a>'

    @scenario
    def a2(self):
        self.widget = A(self.view, Url('/xyz'), description='description')
        self.expected_html = '<a href="/xyz">description</a>'

    @scenario
    def a3(self):
        def disallowed(): return False
        self.widget = A(self.view, Url('/xyz'), description='description', write_check=disallowed)
        self.expected_html = '<a>description</a>'

    @scenario
    def h(self):
        self.widget = H(self.view, 2)
        self.expected_html = '<h2></h2>'

    @scenario
    def p1(self):
        self.widget = P(self.view)
        self.expected_html = '<p></p>'

    @scenario
    def p2(self):
        self.widget = P(self.view, text='text')
        self.expected_html = '<p>text</p>'

    @scenario
    def p_with_slots(self):
        template_p = P(self.view, text='the {0} {{brown}} {slot2} jumps')
        self.widget = template_p.format(Span(self.view, text='quick'), slot2=Span(self.view, text='fox'))
        self.expected_html = '<p>the <span>quick</span> {brown} <span>fox</span> jumps</p>'

    @scenario
    def title(self):
        self.widget = Title(self.view, 'text')
        self.expected_html = '<title>text</title>'

    @scenario
    def link(self):
        self.widget = Link(self.view, 'rr', 'hh', 'tt')
        self.expected_html = '<link href="hh" rel="rr" type="tt">'

    @scenario
    def header(self):
        self.widget = Header(self.view)
        self.expected_html = '<header></header>'

    @scenario
    def footer(self):
        self.widget = Footer(self.view)
        self.expected_html = '<footer></footer>'

    @scenario
    def li(self):
        self.widget = Li(self.view)
        self.expected_html = '<li></li>'

    @scenario
    def ul(self):
        self.widget = Ul(self.view)
        self.expected_html = '<ul></ul>'

    @scenario
    def img1(self):
        self.widget = Img(self.view, 'ss')
        self.expected_html = '<img src="ss">'

    @scenario
    def img2(self):
        self.widget = Img(self.view, 'ss', alt='aa')
        self.expected_html = '<img alt="aa" src="ss">'

    @scenario
    def span(self):
        self.widget = Span(self.view)
        self.expected_html = '<span></span>'

    @scenario
    def span_with_text(self):
        self.widget = Span(self.view, text='some text')
        self.expected_html = '<span>some text</span>'

    @scenario
    def div(self):
        self.widget = Div(self.view)
        self.expected_html = '<div></div>'

    @scenario
    def caption(self):
        self.widget = Caption(self.view, text='some text')
        self.expected_html = '<caption>some text</caption>'

    @scenario
    def col(self):
        self.widget = Col(self.view, span='2')
        self.expected_html = '<col span="2">'

    @scenario
    def colgroup(self):
        self.widget = ColGroup(self.view, span='2')
        self.expected_html = '<colgroup span="2"></colgroup>'

    @scenario
    def thead(self):
        self.widget = Thead(self.view)
        self.expected_html = '<thead></thead>'

    @scenario
    def tfoot(self):
        self.widget = Tfoot(self.view)
        self.expected_html = '<tfoot></tfoot>'

    @scenario
    def tbody(self):
        self.widget = Tbody(self.view)
        self.expected_html = '<tbody></tbody>'

    @scenario
    def tr(self):
        self.widget = Tr(self.view)
        self.expected_html = '<tr></tr>'

    @scenario
    def th(self):
        self.widget = Th(self.view)
        self.expected_html = '<th></th>'

    @scenario
    def td(self):
        self.widget = Td(self.view)
        self.expected_html = '<td></td>'

    @scenario
    def table(self):
        self.widget = Table(self.view, caption_text='my caption', summary='my summary')
        self.expected_html = '<table summary="my summary"><caption>my caption</caption></table>'

    @scenario
    def nav(self):
        self.widget = Nav(self.view)
        self.expected_html = '<nav></nav>'

    @scenario
    def article(self):
        self.widget = Article(self.view)
        self.expected_html = '<article></article>'

    @scenario
    def label1(self):
        self.widget = Label(self.view)
        self.expected_html = '<label></label>'

    @scenario
    def label2(self):
        self.widget = Label(self.view, text='text')
        self.expected_html = '<label>text</label>'

    @scenario
    def fieldset1(self):
        self.widget = FieldSet(self.view)
        self.expected_html = '<fieldset></fieldset>'

    @scenario
    def fieldset2(self):
        self.widget = FieldSet(self.view, legend_text='text')
        self.expected_html = '<fieldset><legend>text</legend></fieldset>'


@with_fixtures(WidgetScenarios)
def test_basic_html_widgets(widget_scenarios):
    """Several basic widgets merely correspond to html elements."""

    tester = WidgetTester(widget_scenarios.widget)
    rendered_html = tester.render_html()
    assert rendered_html == widget_scenarios.expected_html


@with_fixtures(WebFixture)
def test_view_rights_propagate_to_a(web_fixture):
    """The access rights specified for a View are propagated to an A, made from a Bookmark to that View."""
    fixture = web_fixture

    fixture.view.write_check = EmptyStub()
    fixture.view.read_check = EmptyStub()
    a = A.from_bookmark(fixture.view, fixture.view.as_bookmark())
    assert a.read_check is fixture.view.read_check
    assert a.write_check is fixture.view.write_check


@with_fixtures(WebFixture)
def test_text_node_can_vary(web_fixture):
    """A TextNode can vary its text if constructed with a getter for the value instead of a hardcoded value."""

    fixture = web_fixture

    def getter():
        return fixture.current_value


    widget = TextNode(fixture.view, getter)
    tester = WidgetTester(widget)

    fixture.current_value = 'stuff'
    rendered = tester.render_html()
    assert rendered == 'stuff'

    fixture.current_value = 'other'
    rendered = tester.render_html()
    assert rendered == 'other'


@with_fixtures(WebFixture)
def test_text_node_escapes_html(web_fixture):
    """The text of a TextNode is html-escaped."""

    fixture = web_fixture

    widget = TextNode(fixture.view, '<tag> "Here" & \'there\'')
    tester = WidgetTester(widget)

    rendered = tester.render_html()
    assert rendered == '&lt;tag&gt; "Here" &amp; \'there\''


@with_fixtures(WebFixture)
def test_literal_html(web_fixture):
    """The LiteralHTML Widget just renders a chunk of HTML, but can answer queries about images in that HTML."""

    fixture = web_fixture

    contents = '<img src="_some_images/piet.pdf  "> <img src   = \' _some_images/koos was-^-hoêr.jpg\'>'
    literal_html = LiteralHTML(fixture.view, contents)
    tester = WidgetTester(literal_html)

    rendered_html = tester.render_html()
    assert rendered_html == contents

    # Case: when the content is transformed
    def text_transformation(text):
        return text.replace('im', 'IM')
    literal_html = LiteralHTML(fixture.view, contents, transform=text_transformation)
    tester = WidgetTester(literal_html)

    rendered_html = tester.render_html()
    assert rendered_html == text_transformation(contents)


@with_fixtures(WebFixture)
def test_head(web_fixture):
    """Head corresponds with the head HTML element, can have a title and always has a special Slot used by the framework."""

    fixture = web_fixture

    head = Head(fixture.view, 'a title')
    tester = WidgetTester(head)

    reahl_header_slot = head.children[1]
    assert isinstance(reahl_header_slot, Slot)
    assert reahl_header_slot.name == 'reahl_header'

    rendered_html = tester.render_html()
    assert rendered_html == '<head><title>a title</title></head>'


@with_fixtures(WebFixture)
def test_body(web_fixture):
    """Body corresponds with the body HTML element, and always has a special Slot at its end used by the framework."""

    fixture = web_fixture

    body = Body(fixture.view)
    tester = WidgetTester(body)

    body.add_child(P(fixture.view))

    reahl_footer_slot = body.children[-1]
    assert isinstance(reahl_footer_slot, Slot)
    assert reahl_footer_slot.name == 'reahl_footer'

    rendered_html = tester.render_html()
    assert rendered_html == '<body><div id="_reahl_out_of_bound_container"></div><p></p></body>'


@with_fixtures(WebFixture)
def test_html5_page(web_fixture):
    """An HTML5Page is an empty HTML 5 document using the header and body widgets."""

    fixture = web_fixture

    widget = HTML5Page(fixture.view, title='It: $current_title')
    widget.add_default_slot('slot1', P.factory())
    tester = WidgetTester(widget)

    rendered_html = tester.render_html()
    head = '<head><title>It: %s</title></head>' % fixture.view.title
    expected_regex = '^<!DOCTYPE html><html.*class="no-js"><script>.*</script>%s<body><div id="_reahl_out_of_bound_container"></div></body></html>$' % head
    assert re.match(expected_regex, rendered_html.replace('\n', ''))

    assert list(widget.default_slot_definitions.keys()) == ['slot1']



# Copyright 2015-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



import webob

from reahl.tofu import Fixture
from reahl.tofu.pytestsupport import with_fixtures

from reahl.web.fw import Widget
from reahl.web.ui import Img
from reahl.web.bootstrap.carousel import Carousel

from reahl.web_dev.fixtures import WebFixture
from reahl.component_dev.test_i18n import LocaleContextStub


class CarouselFixture(Fixture):
    def get_main_div_for(self, carousel):
        return carousel.children[0]

    def get_indicator_list_for(self, carousel):
        ol = self.get_main_div_for(carousel).children[0]
        assert self.is_carousel_indicator(ol)
        return ol.children

    def is_carousel_indicator(self, element):
        return 'carousel-indicators' in element.get_attribute('class')

    def carousel_indicators_present(self, carousel):
        return any([self.is_carousel_indicator(child)
                    for child in self.get_main_div_for(carousel).children])


@with_fixtures(WebFixture)
def test_carousel_basics(web_fixture):
    """A Carousel contains all the right classes and contents to act as a Bootstrap Carousel component."""

    widget = Carousel(web_fixture.view, 'my_carousel_id')

    [main_div] = widget.children

    # The main div
    assert main_div.get_attribute('id') == 'my_carousel_id'
    assert main_div.get_attribute('class') == 'carousel slide'

    [indicator_list, carousel_inner, left_control, right_control] = main_div.children

    # Indicators
    assert indicator_list.get_attribute('class') == 'carousel-indicators'

    # Inner (container of the images)
    assert carousel_inner.get_attribute('class') == 'carousel-inner'

    # Controls
    def check_control(control, action, label):
        assert control.get_attribute('class') == 'carousel-control-%s' % action
        assert control.get_attribute('role') == 'button'
        assert control.get_attribute('data-slide') == action

        [icon, text] = control.children
        assert icon.get_attribute('class') == 'carousel-control-%s-icon' % action
        assert icon.get_attribute('aria-hidden') == 'true'

        assert text.children[0].value == label
        assert text.get_attribute('class') == 'sr-only'

    check_control(left_control, 'prev', 'Previous')
    check_control(right_control, 'next', 'Next')


@with_fixtures(WebFixture)
def test_i18n(web_fixture):
    """User-visible labels are internationalised."""

    with LocaleContextStub(locale='af') as context:
        context.config = web_fixture.config
        context.session = web_fixture.context.session
        context.request = webob.Request.blank('/', charset='utf8')

        widget = Carousel(web_fixture.view, 'my_carousel_id')

        [main_div] = widget.children
        [indicator_list, carousel_inner, left_control, right_control] = main_div.children

        def check_control(control, label):
            [icon, text] = control.children
            assert text.children[0].value == label

        check_control(left_control, 'Vorige')
        check_control(right_control, 'Volgende')


@with_fixtures(WebFixture, CarouselFixture)
def test_carousel_has_options(web_fixture, carousel_fixture):
    """Constructor allows you to set certain customizing options"""

    carousel = Carousel(web_fixture.view, 'my_carousel_id', interval=1000, pause='hover', wrap=True, keyboard=True)
    main_div = carousel_fixture.get_main_div_for(carousel)

    assert main_div.get_attribute('data-interval') == '1000'
    assert main_div.get_attribute('data-pause') == 'hover'
    assert main_div.get_attribute('data-wrap') == 'true'
    assert main_div.get_attribute('data-keyboard') == 'true'


@with_fixtures(WebFixture, CarouselFixture)
def test_adding_items_to_carousel(web_fixture, carousel_fixture):
    """Images can be added to a Carousel."""

    fixture = carousel_fixture

    carousel = Carousel(web_fixture.view, 'my_carousel_id', show_indicators=True)
    main_div = fixture.get_main_div_for(carousel)
    [indicator_list, carousel_inner, left_control, right_control] = main_div.children

    # Initially, no items or indicators are present
    assert carousel_inner.children == []
    assert fixture.carousel_indicators_present(carousel)
    assert fixture.get_indicator_list_for(carousel) == []

    image = Img(web_fixture.view)
    added_item = carousel.add_slide(image)

    # A carousel item was added for the image
    [carousel_item] = carousel_inner.children
    assert carousel_item is added_item
    assert 'carousel-item' in carousel_item.get_attribute('class')

    [actual_image] = carousel_item.children
    assert actual_image is image

    # An indicator was added for the item
    [indicator] = fixture.get_indicator_list_for(carousel)

    assert indicator.get_attribute('data-target') == '#my_carousel_id'
    assert indicator.get_attribute('data-slide-to') == '0'


@with_fixtures(WebFixture, CarouselFixture)
def test_active_state_of_items(web_fixture, carousel_fixture):
    """The first item added is marked as active, and also its corresponding indicator."""

    fixture = carousel_fixture

    carousel = Carousel(web_fixture.view, 'my_carousel_id')
    carousel.add_slide(Img(web_fixture.view))
    carousel.add_slide(Img(web_fixture.view))

    main_div = fixture.get_main_div_for(carousel)
    [indicator_list, carousel_inner, left_control, right_control] = main_div.children

    #only the first item is active
    [carousel_item_1, carousel_item_2] = carousel_inner.children
    assert carousel_item_1.get_attribute('class') == 'active carousel-item'
    assert carousel_item_2.get_attribute('class') == 'carousel-item'

    #only the first indicator is active
    [indicator_0, indicator_1] = fixture.get_indicator_list_for(carousel)

    assert indicator_0.get_attribute('class') == 'active'
    assert not indicator_1.has_attribute('class')
    assert indicator_0.get_attribute('data-slide-to') == '0'
    assert indicator_1.get_attribute('data-slide-to') == '1'



@with_fixtures(WebFixture, CarouselFixture)
def test_item_indicators_are_optional(web_fixture, carousel_fixture):
    """With show_indicators=False, indicators are not added when adding items."""
    fixture = carousel_fixture

    carousel = Carousel(web_fixture.view, 'my_carousel_id', show_indicators=False)

    assert not fixture.carousel_indicators_present(carousel)

    carousel.add_slide(Img(web_fixture.view))

    #after adding the item, indicators shouldn't appear
    assert not fixture.carousel_indicators_present(carousel)



@with_fixtures(WebFixture)
def test_adding_items_with_captions(web_fixture):
    """A Widget can be supplied to be used caption for an added image."""

    carousel = Carousel(web_fixture.view, 'my_carousel_id')

    caption_widget = Widget(web_fixture.view)
    carousel_item = carousel.add_slide(Img(web_fixture.view), caption_widget=caption_widget)

    [image, div_containing_caption] = carousel_item.children
    assert div_containing_caption.get_attribute('class') == 'carousel-caption'

    [actual_caption_widget] = div_containing_caption.children
    assert actual_caption_widget is caption_widget


@with_fixtures(WebFixture)
def test_check_classes_added_for_images(web_fixture):
    """A Widget can be supplied to be used caption for an added image."""
    carousel = Carousel(web_fixture.view, 'my_carousel_id')

    img_widget = Img(web_fixture.view)
    carousel.add_slide(img_widget)
    assert img_widget.get_attribute('class') == 'd-block w-100'



# Copyright 2015-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from reahl.tofu import Fixture, expected, scenario, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.browsertools.browsertools import Browser

from reahl.component.exceptions import IsInstance, ProgrammerError
from reahl.web.fw import UserInterface, Layout
from reahl.web.ui import P, HTML5Page, Div, Header, Footer
from reahl.web.layout import PageLayout

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_widget_layout(web_fixture):
    """A Layout is used to add children to the Widget in customised ways, and to customise the Widget itself upon construction."""

    class MyLayout(Layout):
        def customise_widget(self):
            self.widget.append_class('class-added-by-custom-layout')

        def add_wrapped(self, child):
            wrapper = self.widget.add_child(Div(self.view))
            wrapper.add_child(child)
            return child

    fixture = web_fixture

    widget_with_layout = Div(fixture.view)

    assert not widget_with_layout.has_attribute('class')
    assert not widget_with_layout.children

    widget_with_layout.use_layout(MyLayout())

    assert widget_with_layout.get_attribute('class') == 'class-added-by-custom-layout'
    assert not widget_with_layout.children

    widget_to_add = P(fixture.view)
    widget_with_layout.layout.add_wrapped(widget_to_add)

    [wrapper] = widget_with_layout.children
    assert wrapper.children == [widget_to_add]

    
@with_fixtures(WebFixture)
def test_widget_layout_errors(web_fixture):
    """A Layout can only be used with a single Widget, and a Widget can only have a single Layout."""

    fixture = web_fixture

    widget_with_layout = Div(fixture.view).use_layout(Layout())

    with expected(ProgrammerError):
        widget_with_layout.use_layout(Layout())

    re_used_layout = Layout()
    widget_with_reused_layout = Div(fixture.view).use_layout(re_used_layout)
    with expected(ProgrammerError):
        Div(fixture.view).use_layout(re_used_layout)


@uses(web_fixture=WebFixture)
class WidgetCreationScenarios(Fixture):

    @property
    def view(self):
        return self.web_fixture.view

    def new_MyLayout(self):
        class MyLayout(Layout):
            def customise_widget(self):
                self.widget.add_child(P(self.view, text='This widget is using Mylayout'))
        return MyLayout

    @scenario
    def use_layout_with_factory_class_method(self):
        fixture = self
        class MainUI(UserInterface):
            def assemble(self):
                self.define_view('/', title='Hello', page=HTML5Page.factory().use_layout(fixture.MyLayout()))
        self.MainUI = MainUI
        
    @scenario
    def use_layout_with_define_page(self):
        fixture = self
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(fixture.MyLayout())
                self.define_view('/', title='Hello')
        self.MainUI = MainUI


@with_fixtures(WebFixture, WidgetCreationScenarios)
def test_widget_factory_creates_widget_with_layout(web_fixture, widget_creation_scenarios):
    """A Layout can be specified to any WidgetFactory or to UserInterface.define_page"""

    class MyLayout(Layout):
        def customise_widget(self):
            self.widget.add_child(P(self.view, text='This widget is using Mylayout'))

    layout_for_widget = MyLayout()

    class MainUI(UserInterface):
        def assemble(self):
            self.define_view('/', title='Hello', page=HTML5Page.factory(use_layout=layout_for_widget))

    fixture = widget_creation_scenarios

    wsgi_app = web_fixture.new_wsgi_app(site_root=fixture.MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')
    [p] = browser.lxml_html.xpath('//p')
    assert p.text == 'This widget is using Mylayout'


@with_fixtures(WebFixture)
def test_page_layout_basics(web_fixture):
    """A PageLayout adds a Div to the body of its page (the page's document), containing a header, footer 
       with a div inbetween the two for page contents."""

    fixture = web_fixture

    layout = PageLayout()
    widget = HTML5Page(fixture.view).use_layout(layout)

    assert [layout.document] == widget.body.children[1:-1]
    header, contents_div, footer = layout.document.children

    assert isinstance(header, Header)
    assert isinstance(contents_div, Div)
    assert isinstance(footer, Footer)


@with_fixtures(WebFixture)
def test_header_and_footer_slots(web_fixture):
    """PageLayout adds a Slot for Header and Footer."""

    fixture = web_fixture

    page = HTML5Page(fixture.view).use_layout(PageLayout())

    header, contents_div, footer = page.layout.document.children

    assert 'header' in header.available_slots
    assert 'footer' in footer.available_slots


@with_fixtures(WebFixture)
def test_page_layout_content_layout(web_fixture):
    """A PageLayout can be given a Layout it should use to lay out its contents Div."""

    fixture = web_fixture

    contents_layout = Layout()
    widget = HTML5Page(fixture.view).use_layout(PageLayout(contents_layout=contents_layout))

    assert widget.layout.contents.layout is contents_layout


@with_fixtures(WebFixture)
def test_page_layout_only_meant_for_html5page(web_fixture):
    """When an attempting to use a PageLayout on something other than an HTML5Page, a useful exception is raised."""

    fixture = web_fixture

    with expected(IsInstance):
        Div(fixture.view).use_layout(PageLayout())


@with_fixtures(WebFixture)
def test_page_layout_convenience_features(web_fixture):
    """A PageLayout exposes useful methods to get to its contents, and adds ids to certain elements for convenience in CSS."""

    fixture = web_fixture

    layout = PageLayout()
    widget = HTML5Page(fixture.view).use_layout(layout)
    header, contents_div, footer = layout.document.children

    assert layout.document.css_id == 'doc'
    assert header.css_id == 'hd'
    assert footer.css_id == 'ft'
    assert contents_div.css_id == 'bd'
    assert contents_div.get_attribute('role') == 'main'

    assert layout.header is header
    assert layout.contents is contents_div
    assert layout.footer is footer



# Copyright 2016-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2


Styled Inputs that allow a user to choose or upload files.

"""


from reahl.component.i18n import Catalogue
from reahl.component.modelinterface import ExposedNames, Action, Event, Field, UploadedFile
from reahl.component.context import ExecutionContext
import reahl.web.ui
from reahl.web.ui import UniqueFilesConstraint
from reahl.web.bootstrap.ui import Div, Span, Li, Ul
from reahl.web.bootstrap.forms import Button, NestedForm, FormLayout, Label


_ = Catalogue('reahl-web')


class _UnstyledHTMLFileInput(reahl.web.ui.SimpleFileInput):
    def __init__(self, form, bound_field):
        super().__init__(form, bound_field)
        self.append_class('form-control-file')


class FileInputButton(reahl.web.ui.WrappedInput):
    """A single button which activates the browser's file choice dialog
       when clicked. The chosen file will only be uploaded once the
       user clicks on any :class:`Button` associated with the same
       :class:`Form` as this Input.

       :param form: (See :class:`~reahl.web.ui.Input`)
       :param bound_field: (See :class:`~reahl.web.ui.Input`, must be of 
              type :class:`reahl.component.modelinterface.FileField`)

    """
    def __init__(self, form, bound_field):
        label = Label(form.view)
        self.simple_input = label.add_child(_UnstyledHTMLFileInput(form, bound_field))
        self.simple_input.html_representation.append_class('btn-secondary')
        label.add_child(Span(form.view, text=_('Choose file(s)')))
        super().__init__(self.simple_input)
        self.add_child(label)

        label.append_class('reahl-bootstrapfileinputbutton')
        label.append_class('btn')
        label.append_class('btn-primary')
        self.set_html_representation(label)

    @property
    def html_control(self):
        return self.simple_input.html_control

    def get_js(self, context=None):
        js = ['$(".reahl-bootstrapfileinputbutton").bootstrapfileinputbutton({});']
        return super().get_js(context=context) + js


class FileInput(reahl.web.ui.WrappedInput):
    """A visual combination of a two buttons and a status area. When the
    user clicks on the 'Choose file' button, the browser's file choice
    dialog is activated. Once chosen the file name that was chosen is
    shown in the status area. The last button will upload this file
    when clicked (it is automatically clicked if the user's JavaScript
    is enabled).

    :param form: (See :class:`~reahl.web.ui.Input`)
    :param bound_field: (See :class:`~reahl.web.ui.Input`, must be of 
           type :class:`reahl.component.modelinterface.FileField`)

    """
    def __init__(self, form, bound_field):
        file_input = FileInputButton(form, bound_field)
        super().__init__(file_input)

        self.input_group = self.add_child(Div(self.view))
        self.input_group.append_class('input-group')
        self.input_group.append_class('reahl-bootstrapfileinput')
        self.set_html_representation(self.input_group)

        div = self.input_group.add_child(Div(form.view))
        div.append_class('input-group-prepend')
        div.add_child(file_input)

        filename_input = self.input_group.add_child(Span(self.view, text=_('No files chosen')))
        filename_input.append_class('form-control')


    def get_js(self, context=None):
        js = ['$(".reahl-bootstrapfileinput").bootstrapfileinput({nfilesMessage: "%s", nofilesMessage: "%s"});' % \
              (_('files chosen'), _('No files chosen'))]
        return super().get_js(context=context) + js



# Uses: reahl/web/reahl.files.js
class FileUploadLi(Li):
    def __init__(self, form, remove_event, persisted_file, css_id=None):
        super().__init__(form.view, css_id=css_id)
        self.set_attribute('class', 'reahl-bootstrap-file-upload-li')
        self.add_child(Button(form, remove_event.with_arguments(filename=persisted_file.filename), style='secondary'))
        self.add_child(Span(self.view, persisted_file.filename))

    def get_js(self, context=None):
        return ['$(".reahl-bootstrap-file-upload-li").bootstrapfileuploadli();']



# Uses: reahl/web/reahl.files.js
class FileUploadPanel(Div):
    def __init__(self, file_upload_input, css_id=None):
        super().__init__(file_upload_input.view, css_id=css_id)
        self.set_attribute('class', 'reahl-bootstrap-file-upload-panel')
        self.file_upload_input = file_upload_input

        self.add_nested_form()
        self.add_uploaded_list()
        self.add_upload_controls()

    def add_nested_form(self):
        self.upload_form = self.add_child(NestedForm(self.view, '%s-%s-upload' % (self.input_form.css_id, self.bound_field.name)))
        self.upload_form.define_event_handler(self.events.upload_file)
        self.upload_form.define_event_handler(self.events.remove_file)
        self.upload_form.form.set_attribute('novalidate','novalidate')
        return self.upload_form

    @property
    def name(self):
        return self.bound_field.name_in_input

    def add_uploaded_list(self):
        ul = self.upload_form.add_child(Ul(self.view))
        for persisted_file in self.persisted_file_class.get_persisted_for_form(self.input_form, self.name):
            ul.add_child(FileUploadLi(self.upload_form.form, self.events.remove_file, persisted_file))
        return ul

    def add_upload_controls(self):
        controls_panel = self.upload_form.add_child(Div(self.view)).use_layout(FormLayout())
        file_input = controls_panel.layout.add_input(FileInput(self.upload_form.form, self.fields.uploaded_file), hide_label=True)
        
        button_addon = file_input.html_representation.add_child(Div(self.view))
        button_addon.append_class('input-group-append')
        button_addon.add_child(Button(self.upload_form.form, self.events.upload_file, style='secondary', outline=True))
        return controls_panel

    @property
    def persisted_file_class(self):
        return self.file_upload_input.persisted_file_class

    @property
    def input_form(self):
        return self.file_upload_input.form

    @property
    def bound_field(self):
        return self.file_upload_input.bound_field

    events = ExposedNames()
    events.upload_file = lambda i: Event(label=_('Upload'), action=Action(i.upload_file))
    events.remove_file = lambda i: Event(label=_('Remove'),
                                         action=Action(i.remove_file, ['filename']),
                                         filename=Field(required=True))

    fields = ExposedNames()
    def create_uploaded_file_field(i):
        field = i.bound_field.unbound_copy()
        field.disallow_multiple()
        field.label = _('Add file')
        field.make_optional()
        field.add_validation_constraint(UniqueFilesConstraint(i.input_form, i.bound_field.name_in_input))
        return field
    fields.uploaded_file = create_uploaded_file_field

    def attach_jq_widget(self, selector, widget_name, **options):
        def js_repr(value):
            if isinstance(value, str):
                return '"%s"' % value
            return value
        option_args = ','.join(['%s: %s' % (name, js_repr(value)) for name, value in options.items()])
        return '$(%s).%s({%s});' % (selector, widget_name, option_args)

    def get_js(self, context=None):
        selector = self.contextualise_selector('".reahl-bootstrap-file-upload-panel"', context)
        unique_names_constraint = self.fields.uploaded_file.get_validation_constraint_of_class(UniqueFilesConstraint)
        js = self.attach_jq_widget(selector, 'bootstrapfileuploadpanel',
                    form_id=self.upload_form.form.css_id,
                    nested_form_id=self.upload_form.css_id,
                    input_form_id=self.input_form.css_id,
                    errorMessage=_('an error occurred, please try again later.'),
                    removeLabel=self.events.remove_file.label,
                    cancelLabel=_('Cancel'),
                    duplicateValidationErrorMessage=unique_names_constraint.message,
                    waitForUploadsMessage=_('Please try again when all files have finished uploading.'))
        return super().get_js(context=context) + [js]

    def remove_file(self, filename):
        self.persisted_file_class.remove_persisted_for_form(self.input_form, self.name, filename)

    def upload_file(self):
        if self.uploaded_file is not None:
            self.persisted_file_class.add_persisted_for_form(self.input_form, self.name, self.uploaded_file)


class FileUploadInput(reahl.web.ui.Input):
    """A Widget that allows the user to upload several files.  FileUploadInput
    makes use of JavaScript to save a user some time: once you choose a file, 
    it is immediately uploaded to the server in the background so that you can
    continue choosing more files.

    Controls are provided so you can cancel uploads that are in
    progress or remove ones that have finished. While a file is
    uploading a progress bar is also shown.

    :param form: (See :class:`~reahl.web.ui.Input`)
    :param bound_field: (See :class:`~reahl.web.ui.Input`, must be of 
              type :class:`reahl.component.modelinterface.FileField`)
    :keyword ignore_concurrent_change: If True, don't check for possible concurrent changes by others to this input (just override such changes).


    .. versionchanged:: 5.0
       Subclass of :class:`~reahl.web.ui.Input` and not :class:`~reahl.web.ui.PrimitiveInput`

    .. versionchanged:: 5.0
       Added `ignore_concurrency_change`

    """
    is_for_file = False

    def __init__(self, form, bound_field, ignore_concurrency_change=False):
        super().__init__(form, bound_field)
        
        self.ignore_concurrency_change = ignore_concurrency_change

        form.register_input(self) # bound_field must be set for this registration to work

        self.set_html_representation(self.add_child(self.create_html_widget()))


    def get_concurrency_hash_strings(self):
        if not self.ignore_concurrency_change:
            yield self.original_database_value
            
    @property
    def name(self):
        return self.bound_field.name_in_input

    def get_ocurred_event(self):
        return None

    @property
    def persisted_file_class(self):
        config = ExecutionContext.get_context().config
        return config.web.persisted_file_class

    def create_html_widget(self):
        return FileUploadPanel(self)

    @property
    def html_control(self):
        return None

    def accept_input(self, input_values):
        value = [UploadedFile(f.filename, f.file_obj.read(), f.mime_type)
                 for f in self.persisted_file_class.get_persisted_for_form(self.form, self.name)]
        self.bound_field.from_input(value)

    def persist_input(self, input_values):
        pass



# Copyright 2014-2023 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.



from reahl.tofu.pytestsupport import with_fixtures

from reahl.web.bootstrap.ui import Alert, Badge, P

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_alerts(web_fixture):
    """An alert is used to display a message with a specific severity"""


    alert = Alert(web_fixture.view, 'Be careful', 'danger')

    assert alert.get_attribute('class') == 'alert alert-danger'
    assert alert.get_attribute('role') == 'alert'
    [message] = alert.children
    assert message.value == 'Be careful'


@with_fixtures(WebFixture)
def test_alert_with_widget_as_message(web_fixture):
    """An alert can also be constructed with a widget as message"""

    widget_contained_in_alert = P(web_fixture.view, text='Consider yourself warned')
    alert = Alert(web_fixture.view, widget_contained_in_alert, 'warning')

    [paragraph] = alert.children
    [message] = paragraph.children
    assert message.value == 'Consider yourself warned'


@with_fixtures(WebFixture)
def test_badges(web_fixture):
    """A badge is used inside other widgets to draw the users attention"""

    badge = Badge(web_fixture.view, 'On sale', 'info')

    assert badge.get_attribute('class') == 'badge badge-info'
    [message] = badge.children
    assert message.value == 'On sale'


@with_fixtures(WebFixture)
def test_badge_as_pill(web_fixture):
    """A badge can be made to look like a pill"""

    badge = Badge(web_fixture.view, 'On sale', 'success', pill=True)

    assert badge.get_attribute('class') == 'badge badge-pill badge-success'
